"""
translate.py - Sopel Translation Plugin
Copyright 2008, Sean B. Palmer, inamidst.com
Copyright 2013-2014, Elad Alfassa <elad@fedoraproject.org>
Licensed under the Eiffel Forum License 2.

https://sopel.chat
"""
from __future__ import annotations

import json
import logging
import random
import re

import requests

from sopel import plugin
from sopel.tools import web


LOGGER = logging.getLogger(__name__)
PLUGIN_OUTPUT_PREFIX = '[translate] '


def setup(bot):
    if 'mangle_lines' not in bot.memory:
        bot.memory['mangle_lines'] = bot.make_identifier_memory()


def shutdown(bot):
    try:
        del bot.memory['mangle_lines']
    except KeyError:
        pass


def translate(text, in_lang='auto', out_lang='en'):
    raw = False
    if str(out_lang).endswith('-raw'):
        out_lang = out_lang[:-4]
        raw = True

    headers = {
        'User-Agent': 'Mozilla/5.0' +
        '(X11; U; Linux i686)' +
        'Gecko/20071127 Firefox/2.0.0.11'
    }

    query = {
        "client": "gtx",
        "sl": in_lang,
        "tl": out_lang,
        "dt": "t",
        "q": text,
    }
    url = "https://translate.googleapis.com/translate_a/single"
    result = requests.get(url, params=query, timeout=40, headers=headers).text

    if result == '[,,""]':
        return None, in_lang

    while ',,' in result:
        result = result.replace(',,', ',null,')
        result = result.replace('[,', '[null,')

    try:
        data = json.loads(result)
    except ValueError:
        LOGGER.error(
            'Error parsing JSON response from translate API (%s to %s: "%s")',
            in_lang, out_lang, text)
        return None, None

    if raw:
        return str(data), 'en-raw'

    try:
        language = data[2]  # -2][0][0]
    except IndexError:
        language = '?'

    return ''.join(x[0] for x in data[0]), language


@plugin.rule(r'$nickname[,:]\s+(?:([a-z]{2}) +)?(?:([a-z]{2}|en-raw) +)?["“](.+?)["”]\? *$')
@plugin.example('$nickname: "mon chien"? or $nickname: fr "mon chien"?')
@plugin.priority('low')
@plugin.output_prefix(PLUGIN_OUTPUT_PREFIX)
def tr(bot, trigger):
    """Translates a phrase, with an optional language hint."""
    in_lang, out_lang, phrase = trigger.groups()

    if (len(phrase) > 350) and (not trigger.admin):
        bot.reply('Phrase must be under 350 characters.')
        return

    if phrase.strip() == '':
        bot.reply('You need to specify a string for me to translate!')
        return

    in_lang = in_lang or 'auto'
    out_lang = out_lang or 'en'

    if in_lang == out_lang:
        bot.reply('Language guessing failed, so try suggesting one!')
        return

    try:
        msg, in_lang = translate(phrase, in_lang, out_lang)
    except requests.Timeout:
        bot.reply("Translation service unavailable (timeout).")
        LOGGER.error(
            'Translate API error (%s to %s: "%s"): timeout.',
            in_lang, out_lang, phrase)
        return
    except requests.RequestException as http_error:
        bot.reply("Translation request failed.")
        LOGGER.exception(
            'Translate API error (%s to %s: "%s"): %s.',
            in_lang, out_lang, phrase, http_error)
        return

    if not in_lang:
        bot.reply("Translation failed, probably because of a rate-limit.")
        return

    if not msg:
        bot.reply(
            'The %s to %s translation failed; are you sure you specified '
            'valid language abbreviations?' % (in_lang, out_lang)
        )
        return

    msg = web.decode(msg)
    msg = '"%s" (%s to %s, translate.google.com)' % (msg, in_lang, out_lang)
    bot.say(msg)


@plugin.command('translate', 'tr')
@plugin.example('.tr :en :fr my dog',
                '"mon chien" (en to fr, translate.google.com)',
                online=True, vcr=True, user_help=True)
@plugin.example('.tr מחשב',
                '"computer" (iw to en, translate.google.com)',
                online=True, vcr=True)
@plugin.example('.tr mon chien',
                '"my dog" (fr to en, translate.google.com)',
                online=True, vcr=True, user_help=True)
@plugin.output_prefix(PLUGIN_OUTPUT_PREFIX)
def tr2(bot, trigger):
    """Translates a phrase, with an optional language :hint."""
    command = trigger.group(2)

    if not command:
        bot.reply('You did not give me anything to translate.')
        return

    def langcode(p):
        # TODO: We'd be much better off just using the langcodes PyPI package
        # also TODO: it'd be nice not to require the : prefix, which using
        # langcodes would make easier in lieu of adding an API key to fetch
        # the list of supported languages
        prefixed = p.startswith(':')
        # the longest codes Google uses (ca. Jan 2022) are zh-CN and zh-TW
        short_enough = (2 < len(p) < 8)
        # pesky - in Chinese codes forced a switch from .isalpha(); see #2241
        fits_pattern = re.match(r':[a-z\-]+', p.lower())
        return all((prefixed, short_enough, fits_pattern))

    args = ['auto', 'en']

    for i in range(2):
        if ' ' not in command:
            break
        prefix, cmd = command.split(' ', 1)
        if langcode(prefix):
            args[i] = prefix[1:]
            command = cmd

    phrase = command
    if (len(phrase) > 350) and (not trigger.admin):
        bot.reply('Phrase must be under 350 characters.')
        return

    if phrase.strip() == '':
        bot.reply('You need to specify a string for me to translate!')
        return

    src, dest = args

    if src == dest:
        bot.reply('Language guessing failed, so try suggesting one!')
        return

    try:
        msg, src = translate(phrase, src, dest)
    except requests.Timeout:
        bot.reply("Translation service unavailable (timeout).")
        LOGGER.error(
            'Translate API error (%s to %s: "%s"): timeout.',
            src, dest, phrase)
        return
    except requests.RequestException as http_error:
        bot.reply("Translation request failed.")
        LOGGER.exception(
            'Translate API error (%s to %s: "%s"): %s.',
            src, dest, phrase, http_error)
        return

    if not src:
        return bot.say("Translation failed, probably because of a rate-limit.")

    if not msg:
        bot.reply(
            'The %s to %s translation failed; '
            'are you sure you specified valid language abbreviations?'
            % (src, dest))
        return

    msg = web.decode(msg)  # msg.replace('&#39;', "'")
    msg = '"%s" (%s to %s, translate.google.com)' % (msg, src, dest)

    bot.say(msg)


def get_random_lang(long_list, short_list):
    random_index = random.randint(0, len(long_list) - 1)
    random_lang = long_list[random_index]
    if random_lang not in short_list:
        short_list.append(random_lang)
    else:
        return get_random_lang(long_list, short_list)
    return short_list


@plugin.command('mangle', 'mangle2')
@plugin.output_prefix(PLUGIN_OUTPUT_PREFIX)
def mangle(bot, trigger):
    """Repeatedly translate the input until it makes absolutely no sense."""
    long_lang_list = ['fr', 'de', 'es', 'it', 'no', 'he', 'la', 'ja', 'cy', 'ar', 'yi', 'zh', 'nl', 'ru', 'fi', 'hi', 'af', 'jw', 'mr', 'ceb', 'cs', 'ga', 'sv', 'eo', 'el', 'ms', 'lv']
    lang_list: list[str] = []

    for __ in range(0, 8):
        lang_list = get_random_lang(long_lang_list, lang_list)
    random.shuffle(lang_list)

    phrase: tuple[str, ...] = ()
    if trigger.group(2) is None:
        try:
            phrase = (bot.memory['mangle_lines'][trigger.sender], '')
        except KeyError:
            bot.reply("What do you want me to mangle?")
            return
    else:
        phrase = (trigger.group(2).strip(), '')

    if phrase[0] == '':
        bot.reply("What do you want me to mangle?")
        return

    for lang in lang_list:
        backup = phrase
        try:
            phrase = translate(phrase[0], 'en', lang)
        except Exception:  # TODO: Be specific
            phrase = ()
        if not phrase:
            phrase = backup
            break

        try:
            phrase = translate(phrase[0], lang, 'en')
        except Exception:  # TODO: Be specific
            phrase = backup
            continue

        if not phrase:
            phrase = backup
            break

    if phrase[0] is None:
        # translate() returns (None, None) if an error happens,
        # usually because the bot has exceeded a rate limit
        bot.reply("Translation rate limit reached. Try again later.")
        return

    bot.say(phrase[0])


@plugin.rule('(.*)')
@plugin.priority('low')
@plugin.unblockable
def collect_mangle_lines(bot, trigger):
    bot.memory['mangle_lines'][trigger.sender] = "%s said '%s'" % (
        trigger.nick,
        trigger.group(0).strip(),
    )



"""
dice.py - Sopel Dice Plugin
Copyright 2010-2013, Dimitri "Tyrope" Molenaars, TyRope.nl
Copyright 2013, Ari Koivula, <ari@koivu.la>
Licensed under the Eiffel Forum License 2.

https://sopel.chat
"""
from __future__ import annotations

import operator
import random
import re
from typing import TYPE_CHECKING

from sopel import plugin
from sopel.tools.calculation import eval_equation

if TYPE_CHECKING:
    from sopel.bot import SopelWrapper
    from sopel.trigger import Trigger


MAX_DICE = 1000


class DicePouch:
    def __init__(self, dice_count: int, dice_type: int) -> None:
        """Initialize dice pouch and roll the dice.

        :param dice_count: the number of dice in the pouch
        :param dice_type: how many faces each die has
        """
        self.num: int = dice_count
        self.type: int = dice_type

        self.dice: dict[int, int] = {}
        self.dropped: dict[int, int] = {}

        self.roll_dice()

    def roll_dice(self) -> None:
        """Roll all the dice in the pouch."""
        self.dice = {}
        self.dropped = {}
        for __ in range(self.num):
            number = random.randint(1, self.type)
            count = self.dice.setdefault(number, 0)
            self.dice[number] = count + 1

    def drop_lowest(self, n: int) -> None:
        """Drop ``n`` lowest dice from the result.

        :param n: the number of dice to drop
        """

        sorted_x = sorted(self.dice.items(), key=operator.itemgetter(0))

        for i, count in sorted_x:
            count = self.dice[i]
            if n == 0:
                break
            elif n < count:
                self.dice[i] = count - n
                self.dropped[i] = n
                break
            else:
                self.dice[i] = 0
                self.dropped[i] = count
                n = n - count

        for i, count in self.dropped.items():
            if self.dice[i] == 0:
                del self.dice[i]

    def get_simple_string(self) -> str:
        """Return the values of the dice like (2+2+2[+1+1])."""
        dice = self.dice.items()
        faces = ("+".join([str(face)] * times) for face, times in dice)
        dice_str = "+".join(faces)

        dropped_str = ""
        if self.dropped:
            dropped = self.dropped.items()
            dfaces = ("+".join([str(face)] * times) for face, times in dropped)
            dropped_str = "[+%s]" % ("+".join(dfaces),)

        return "(%s%s)" % (dice_str, dropped_str)

    def get_compressed_string(self) -> str:
        """Return the values of the dice like (3x2[+2x1])."""
        dice = self.dice.items()
        faces = ("%dx%d" % (times, face) for face, times in dice)
        dice_str = "+".join(faces)

        dropped_str = ""
        if self.dropped:
            dropped = self.dropped.items()
            dfaces = ("%dx%d" % (times, face) for face, times in dropped)
            dropped_str = "[+%s]" % ("+".join(dfaces),)

        return "(%s%s)" % (dice_str, dropped_str)

    def get_sum(self) -> int:
        """Get the sum of non-dropped dice."""
        result = 0
        for face, times in self.dice.items():
            result += face * times
        return result

    def get_number_of_faces(self) -> int:
        """Returns sum of different faces for dropped and not dropped dice.

        This can be used to estimate whether the result can be shown (in
        compressed form) in a reasonable amount of space.
        """
        return len(self.dice) + len(self.dropped)


class DiceError(Exception):
    """Custom base exception type."""


class InvalidDiceFacesError(DiceError):
    """Custom exception type for invalid number of die faces."""
    def __init__(self, faces: int):
        super().__init__(faces)

    @property
    def faces(self) -> int:
        return self.args[0]


class NegativeDiceCountError(DiceError):
    """Custom exception type for invalid numbers of dice."""
    def __init__(self, count: int):
        super().__init__(count)

    @property
    def count(self) -> int:
        return self.args[0]


class TooManyDiceError(DiceError):
    """Custom exception type for excessive numbers of dice."""
    def __init__(self, requested: int, available: int):
        super().__init__(requested, available)

    @property
    def available(self) -> int:
        return self.args[1]

    @property
    def requested(self) -> int:
        return self.args[0]


class UnableToDropDiceError(DiceError):
    """Custom exception type for failing to drop lowest N dice."""
    def __init__(self, dropped: int, total: int):
        super().__init__(dropped, total)

    @property
    def dropped(self) -> int:
        return self.args[0]

    @property
    def total(self) -> int:
        return self.args[1]


def _get_error_message(exc: DiceError) -> str:
    if isinstance(exc, InvalidDiceFacesError):
        return "I don't have any dice with {} sides.".format(exc.faces)
    if isinstance(exc, NegativeDiceCountError):
        return "I can't roll {} dice.".format(exc.count)
    if isinstance(exc, TooManyDiceError):
        return "I only have {}/{} dice.".format(exc.available, exc.requested)
    if isinstance(exc, UnableToDropDiceError):
        return "I can't drop the lowest {} of {} dice.".format(
            exc.dropped, exc.total)

    return "Unknown error rolling dice: %r" % exc


def _roll_dice(dice_match: re.Match[str]) -> DicePouch:
    dice_num = int(dice_match.group('dice_num') or 1)
    dice_type = int(dice_match.group('dice_type'))

    # Dice can't have zero or a negative number of sides.
    if dice_type <= 0:
        raise InvalidDiceFacesError(dice_type)

    # Can't roll a negative number of dice.
    if dice_num < 0:
        raise NegativeDiceCountError(dice_num)

    # Upper limit for dice should be at most a million. Creating a dict with
    # more than a million elements already takes a noticeable amount of time
    # on a fast computer and ~55kB of memory.
    if dice_num > MAX_DICE:
        raise TooManyDiceError(dice_num, MAX_DICE)

    dice = DicePouch(dice_num, dice_type)

    if dice_match.group('drop_lowest'):
        drop = int(dice_match.group('drop_lowest'))
        if drop >= 0:
            dice.drop_lowest(drop)
        else:
            raise UnableToDropDiceError(drop, dice_num)

    return dice


@plugin.command('roll', 'dice', 'd')
@plugin.priority("medium")
@plugin.example(".roll", "No dice to roll.")
@plugin.example(".roll 2d6+4^2&",
                "I don't know how to process that. "
                "Are the dice as well as the algorithms correct?")
@plugin.example(".roll 65(2)", "I couldn't find any valid dice expressions.")
@plugin.example(".roll 2d-2", "I don't have any dice with -2 sides.")
@plugin.example(".roll 1d0", "I don't have any dice with 0 sides.")
@plugin.example(".roll -1d6", "I can't roll -1 dice.")
@plugin.example(".roll 3d6v-1", "I can't drop the lowest -1 of 3 dice.")
@plugin.example(".roll 2d6v0", r'2d6v0: \(\d\+\d\) = \d+', re=True)
@plugin.example(".roll 2d6v4", r'2d6v4: \(\[\+\d\+\d\]\) = 0', re=True)
@plugin.example(".roll 2d6v1+8", r'2d6v1\+8: \(\d\[\+\d\]\)\+8 = \d+', re=True)
@plugin.example(".roll 11d1v1", "11d1v1: (10x1[+1x1]) = 10")
@plugin.example(".roll 3d1+1", '3d1+1: (1+1+1)+1 = 4')
@plugin.example(".roll 3d1v2+1", '3d1v2+1: (1[+1+1])+1 = 2')
@plugin.example(".roll 2d4", r'2d4: \(\d\+\d\) = \d', re=True)
@plugin.example(".roll 100d1", r'[^:]*: \(100x1\) = 100', re=True)
@plugin.example(".roll 100d100", r'100d100: \(\.{3}\) = \d+', re=True)
@plugin.example(".roll 1000d999^1000d999", 'You roll 1000d999^1000d999: (...)^(...) = very big')
@plugin.example(".roll 1000d999^1000d99", "I can't display a number that big. =(")
@plugin.example(
    ".roll {}d1".format(MAX_DICE + 1), 'I only have {}/{} dice.'.format(MAX_DICE, MAX_DICE + 1))
@plugin.example(".roll 1d1 + 1d1", '1d1 + 1d1: (1) + (1) = 2')
@plugin.example(".roll 1d1+1d1", '1d1+1d1: (1)+(1) = 2')
@plugin.example(".roll d6 # initiative", r'd6: \(\d\) = \d', re=True)
@plugin.example(".roll 2d20v1+2 # roll with advantage", user_help=True)
@plugin.example(".roll 2d10+3", user_help=True)
@plugin.example(".roll 1d6", user_help=True)
@plugin.output_prefix('[dice] ')
def roll(bot: SopelWrapper, trigger: Trigger) -> None:
    """Rolls dice and reports the result.

    The dice roll follows this format: XdY[vZ][+N][#COMMENT]

    X is the number of dice. Y is the number of faces in the dice. Z is the
    number of lowest dice to be dropped from the result. N is the constant to
    be applied to the end result. Comment is for easily noting the purpose.
    """
    dice_regexp = r"""
        (?P<dice_num>-?\d*)
        d
        (?P<dice_type>-?\d+)
        (v(?P<drop_lowest>-?\d+))?
    """

    # Get a list of all dice expressions, evaluate them and then replace the
    # expressions in the original string with the results. Replacing is done
    # using string formatting, so %-characters must be escaped.
    if not trigger.group(2):
        bot.reply("No dice to roll.")
        return

    arg_str_raw = trigger.group(2).split("#", 1)[0].strip()
    arg_str = arg_str_raw.replace("%", "%%")
    arg_str = re.sub(dice_regexp, "%s", arg_str, 0, re.IGNORECASE | re.VERBOSE)

    dice_expressions = [
        match for match in
        re.finditer(dice_regexp, arg_str_raw, re.IGNORECASE | re.VERBOSE)
    ]

    if not dice_expressions:
        bot.reply("I couldn't find any valid dice expressions.")
        return

    try:
        dice = [_roll_dice(dice_expr) for dice_expr in dice_expressions]
    except DiceError as err:
        # Stop computing roll if there was a problem rolling dice.
        bot.reply(_get_error_message(err))
        return

    def _get_eval_str(dice: DicePouch) -> str:
        return "(%d)" % (dice.get_sum(),)

    def _get_pretty_str(dice: DicePouch) -> str:
        if dice.num <= 10:
            return dice.get_simple_string()
        elif dice.get_number_of_faces() <= 10:
            return dice.get_compressed_string()
        else:
            return "(...)"

    eval_str: str = arg_str % (tuple(map(_get_eval_str, dice)))
    pretty_str: str = arg_str % (tuple(map(_get_pretty_str, dice)))

    try:
        result = eval_equation(eval_str)
    except ValueError:
        # As it seems that ValueError is raised if the resulting equation would
        # be too big, give a semi-serious answer to reflect on this.
        bot.reply("You roll %s: %s = very big" % (
            arg_str_raw, pretty_str))
        return
    except (SyntaxError, eval_equation.Error):
        bot.reply(
            "I don't know how to process that. "
            "Are the dice as well as the algorithms correct?"
        )
        return

    try:
        bot.say("%s: %s = %d" % (arg_str_raw, pretty_str, result))
    except ValueError:
        # Converting the result to a string can also raise ValueError if it has
        # more than int_max_str_digits digits (4300 by default on CPython)
        # See https://docs.python.org/3.12/library/stdtypes.html#int-max-str-digits
        bot.reply("I can't display a number that big. =(")



"""
url.py - Sopel URL Title Plugin
Copyright 2010-2011, Michael Yanovich (yanovich.net) & Kenneth Sham
Copyright 2012-2013, Elsie Powell
Copyright 2013, Lior Ramati <firerogue517@gmail.com>
Copyright 2014, Elad Alfassa <elad@fedoraproject.org>
Licensed under the Eiffel Forum License 2.

https://sopel.chat
"""
from __future__ import annotations

from ipaddress import ip_address
import logging
import re
from typing import NamedTuple, Optional, TYPE_CHECKING
from urllib.parse import urlparse

import dns.resolver
import requests
from urllib3.exceptions import LocationValueError  # type: ignore[import]

from sopel import plugin, privileges, tools
from sopel.config import types
from sopel.tools import web

if TYPE_CHECKING:
    from collections.abc import Generator
    from sopel.bot import Sopel, SopelWrapper
    from sopel.config import Config
    from sopel.trigger import Trigger


LOGGER = logging.getLogger(__name__)
USER_AGENT = (
    'Mozilla/5.0 (Windows NT 10.0; Win64; x64) '
    'AppleWebKit/537.36 (KHTML, like Gecko) '
    'Chrome/98.0.4758.102 Safari/537.36'
)
DEFAULT_HEADERS = {
    'User-Agent': USER_AGENT,
    'Accept': 'text/html, application/xhtml+xml, application/xml;q=0.9, */*;q=0.8',
    'Accept-Language': 'en,en-US;q=0,5',
}
# These are used to clean up the title tag before actually parsing it. Not the
# world's best way to do this, but it'll do for now.
TITLE_TAG_DATA = re.compile('<(/?)title( [^>]+)?>', re.IGNORECASE)
QUOTED_TITLE = re.compile('[\'"]<title>[\'"]', re.IGNORECASE)
# This sets the maximum number of bytes that should be read in order to find
# the title. We don't want it too high, or a link to a big file/stream will
# just keep downloading until there's no more memory. 640k ought to be enough
# for anybody, but the modern web begs to differ.
MAX_BYTES = 655360 * 2


class UrlSection(types.StaticSection):
    enable_auto_title = types.BooleanAttribute(
        'enable_auto_title', default=True)
    """Enable auto-title (enabled by default)"""
    # TODO some validation rules maybe?
    exclude = types.ListAttribute('exclude')
    """A list of regular expressions to match URLs for which the title should not be shown."""
    exclude_required_access = types.ChoiceAttribute(
        'exclude_required_access',
        choices=privileges.__all__,
        default='OP',
    )
    """Minimum channel access level required to edit ``exclude`` list using chat commands."""
    exclusion_char = types.ValidatedAttribute('exclusion_char', default='!')
    """A character (or string) which, when immediately preceding a URL, will stop that URL's title from being shown."""
    shorten_url_length = types.ValidatedAttribute(
        'shorten_url_length', int, default=0)
    """If greater than 0, the title fetcher will include a TinyURL version of links longer than this many characters."""
    enable_private_resolution = types.BooleanAttribute(
        'enable_private_resolution', default=False)
    """Enable requests to private and local network IP addresses"""


def configure(config: Config) -> None:
    """
    | name | example | purpose |
    | ---- | ------- | ------- |
    | enable_auto_title | yes | Enable auto-title. |
    | exclude | https?://git\\\\.io/.* | A list of regular expressions for URLs for which the title should not be shown. |
    | exclusion\\_char | ! | A character (or string) which, when immediately preceding a URL, will stop the URL's title from being shown. |
    | shorten\\_url\\_length | 72 | If greater than 0, the title fetcher will include a TinyURL version of links longer than this many characters. |
    | enable\\_private\\_resolution | False | Enable requests to private and local network IP addresses. |
    """
    config.define_section('url', UrlSection)
    config.url.configure_setting(
        'enable_auto_title',
        'Enable auto-title?'
    )
    config.url.configure_setting(
        'exclude',
        'Enter regular expressions for each URL you would like to exclude.'
    )
    config.url.configure_setting(
        'exclusion_char',
        'Enter a character which can be prefixed to suppress URL titling'
    )
    config.url.configure_setting(
        'shorten_url_length',
        'Enter how many characters a URL should be before the bot puts a'
        ' shorter version of the URL in the title as a TinyURL link'
        ' (0 to disable)'
    )
    config.url.configure_setting(
        'enable_private_resolution',
        'Enable requests to private and local network IP addresses?'
    )


def setup(bot: Sopel) -> None:
    bot.config.define_section('url', UrlSection)

    if bot.config.url.exclude:
        regexes = [re.compile(s) for s in bot.config.url.exclude]
    else:
        regexes = []

    # We're keeping these in their own list, rather than putting then in the
    # callbacks list because 1, it's easier to deal with plugins that are still
    # using this list, and not the newer callbacks list and 2, having a lambda
    # just to pass is kinda ugly.
    if 'url_exclude' not in bot.memory:
        bot.memory['url_exclude'] = regexes
    else:
        exclude = bot.memory['url_exclude']
        if regexes:
            exclude.extend(regexes)
        bot.memory['url_exclude'] = exclude

    # Ensure last_seen_url is in memory
    if 'last_seen_url' not in bot.memory:
        bot.memory['last_seen_url'] = bot.make_identifier_memory()

    # Initialize shortened_urls as a dict if it doesn't exist.
    if 'shortened_urls' not in bot.memory:
        bot.memory['shortened_urls'] = tools.SopelMemory()


def shutdown(bot: Sopel) -> None:
    # Unset `url_exclude` and `last_seen_url`, but not `shortened_urls`;
    # clearing `shortened_urls` will increase API calls. Leaving it in memory
    # should not lead to unexpected behavior.
    for key in ['url_exclude', 'last_seen_url']:
        try:
            del bot.memory[key]
        except KeyError:
            pass


def _user_can_change_excludes(bot: SopelWrapper, trigger: Trigger) -> bool:
    if trigger.admin:
        return True

    required_access = bot.config.url.exclude_required_access
    channel = bot.channels[trigger.sender]
    user_access = channel.privileges[trigger.nick]

    if user_access >= getattr(privileges, required_access):
        return True

    return False


@plugin.command('urlexclude', 'urlpexclude', 'urlban', 'urlpban')
@plugin.example('.urlpexclude example\\.com/\\w+', user_help=True)
@plugin.example('.urlexclude example.com/path', user_help=True)
@plugin.output_prefix('[url] ')
def url_ban(bot: SopelWrapper, trigger: Trigger) -> None:
    """Exclude a URL from auto title.

    Use ``urlpexclude`` to exclude a pattern instead of a URL.
    """
    url = trigger.group(2)

    if not url:
        bot.reply('This command requires a URL to exclude.')
        return

    if not _user_can_change_excludes(bot, trigger):
        bot.reply(
            'Only admins and channel members with %s access or higher may '
            'modify URL excludes.' % bot.config.url.exclude_required_access)
        return

    if trigger.group(1) in ['urlpexclude', 'urlpban']:
        # validate regex pattern
        try:
            re.compile(url)
        except re.error as err:
            bot.reply('Invalid regex pattern: %s' % err)
            return
    else:
        # escape the URL to ensure a valid pattern
        url = re.escape(url)

    patterns = bot.settings.url.exclude

    if url in patterns:
        bot.reply('This URL is already excluded from auto title.')
        return

    # update settings
    patterns.append(url)
    bot.settings.url.exclude = patterns  # set the config option
    bot.settings.save()
    LOGGER.info('%s excluded the URL pattern "%s"', trigger.nick, url)

    # re-compile
    bot.memory['url_exclude'] = [re.compile(s) for s in patterns]

    # tell the user
    bot.reply('This URL is now excluded from auto title.')


@plugin.command('urlallow', 'urlpallow', 'urlunban', 'urlpunban')
@plugin.example('.urlpallow example\\.com/\\w+', user_help=True)
@plugin.example('.urlallow example.com/path', user_help=True)
@plugin.output_prefix('[url] ')
def url_unban(bot: SopelWrapper, trigger: Trigger) -> None:
    """Allow a URL for auto title.

    Use ``urlpallow`` to allow a pattern instead of a URL.
    """
    url = trigger.group(2)

    if not url:
        bot.reply('This command requires a URL to allow.')
        return

    if not _user_can_change_excludes(bot, trigger):
        bot.reply(
            'Only admins and channel members with %s access or higher may '
            'modify URL excludes.' % bot.config.url.exclude_required_access)
        return

    if trigger.group(1) in ['urlpallow', 'urlpunban']:
        # validate regex pattern
        try:
            re.compile(url)
        except re.error as err:
            bot.reply('Invalid regex pattern: %s' % err)
            return
    else:
        # escape the URL to ensure a valid pattern
        url = re.escape(url)

    patterns = bot.settings.url.exclude

    if url not in patterns:
        bot.reply('This URL was not excluded from auto title.')
        return

    # update settings
    patterns.remove(url)
    bot.settings.url.exclude = patterns  # set the config option
    bot.settings.save()
    LOGGER.info('%s allowed the URL pattern "%s"', trigger.nick, url)

    # re-compile
    bot.memory['url_exclude'] = [re.compile(s) for s in patterns]

    # tell the user
    bot.reply('This URL is not excluded from auto title anymore.')


@plugin.command('title')
@plugin.example(
    '.title https://www.google.com',
    'Google | www.google.com',
    online=True, vcr=True)
@plugin.output_prefix('[url] ')
def title_command(bot: SopelWrapper, trigger: Trigger) -> None:
    """
    Show the title or URL information for the given URL, or the last URL seen
    in this channel.
    """
    result_count = 0

    if not trigger.group(2):
        if trigger.sender not in bot.memory['last_seen_url']:
            return
        urls = [bot.memory["last_seen_url"][trigger.sender]]
    else:
        # needs to be a list so len() can be checked later
        urls = list(web.search_urls(trigger))

    for url, title, domain, tinyurl, ignored in process_urls(
        bot, trigger, urls, requested=True
    ):
        if ignored:
            result_count += 1
            continue
        message = "%s | %s" % (title, domain)
        if tinyurl:
            message += ' ( %s )' % tinyurl
        bot.reply(message)
        bot.memory['last_seen_url'][trigger.sender] = url
        result_count += 1

    expected_count = len(urls)
    if result_count < expected_count:
        if expected_count == 1:
            bot.reply("Sorry, fetching that title failed. Make sure the site is working.")
        elif result_count == 0:
            bot.reply("Sorry, I couldn't fetch titles for any of those.")
        else:
            bot.reply("I couldn't get all of the titles, but I fetched what I could!")


@plugin.rule(r'(?u).*(https?://\S+).*')
@plugin.output_prefix('[url] ')
def title_auto(bot: SopelWrapper, trigger: Trigger) -> None:
    """
    Automatically show titles for URLs. For shortened URLs/redirects, find
    where the URL redirects to and show the title for that.

    .. note::

        URLs that match (before redirection) any other registered callbacks
        will *not* have their titles shown.

    """
    # Enabled or disabled by feature flag
    if not bot.settings.url.enable_auto_title:
        return

    # Avoid fetching links from another command
    if re.match(bot.config.core.prefix + r'\S+', trigger):
        return

    unchecked_urls = web.search_urls(
        trigger, exclusion_char=bot.config.url.exclusion_char, clean=True)

    urls = []
    safety_cache = bot.memory.get("safety_cache", {})
    safety_cache_local = bot.memory.get("safety_cache_local", {})
    for url in unchecked_urls:
        # Avoid fetching known malicious links
        if url in safety_cache and safety_cache[url]["positives"] > 0:
            continue
        parsed = urlparse(url)
        if not parsed.hostname or parsed.hostname.lower() in safety_cache_local:
            continue
        urls.append(url)

    for url, title, domain, tinyurl, ignored in process_urls(bot, trigger, urls):
        if not ignored:
            message = '%s | %s' % (title, domain)
            if tinyurl:
                message += ' ( %s )' % tinyurl
            # Guard against responding to other instances of this bot.
            if message != trigger:
                bot.say(message)
        bot.memory["last_seen_url"][trigger.sender] = url


class URLInfo(NamedTuple):
    """Helper class for information about a URL handled by this plugin."""

    url: str

    title: Optional[str]
    """The title associated with ``url``, if appropriate."""

    hostname: Optional[str]
    """The hostname associated with ``url``, if appropriate."""

    tinyurl: Optional[str]
    """A shortened form of ``url``, if appropriate."""

    ignored: bool
    """Whether or not this URL matches any registered callbacks or is explicitly excluded."""


def process_urls(
    bot: SopelWrapper,
    trigger: Trigger,
    urls: list[str],
    requested: bool = False,
) -> Generator[URLInfo, None, None]:
    """
    For each URL in the list, ensure it should be titled, and do so.

    :param bot: Sopel instance
    :param trigger: The trigger object for this event
    :param urls: The URLs detected in the triggering message
    :param requested: Whether the title was explicitly requested (vs automatic)

    Yields a tuple ``(url, title, hostname, tinyurl, ignored)`` for each URL.

    .. note:

        If a URL in ``urls`` has any registered callbacks, this function will NOT
        retrieve the title, and considers the URL as dispatched to those callbacks.
        In this case, only the ``url`` and ``ignored=True`` will be set; all
        other values will be ``None``.

    .. note:

        For titles explicitly requested by the user, ``exclusion_char`` and
        exclusions from the ``.urlban``/``.urlpban`` commands are skipped.

    .. versionchanged:: 8.0

        This function **does not** notify callbacks registered for URLs
        redirected to from URLs passed to this function. See #2432, #2230.

    """
    shorten_url_length = bot.config.url.shorten_url_length
    for url in urls:
        # Exclude URLs that start with the exclusion char
        if not requested and url.startswith(bot.config.url.exclusion_char):
            continue

        parsed_url = urlparse(url)

        if check_callbacks(bot, url, use_excludes=not requested):
            # URL matches a callback OR is excluded, ignore
            yield URLInfo(url, None, None, None, True)
            continue

        # Prevent private addresses from being queried if enable_private_resolution is False
        # FIXME: This does nothing when an attacker knows how to host a 302
        # FIXME: This whole concept has a TOCTOU issue
        if not bot.config.url.enable_private_resolution:
            if not parsed_url.hostname:
                # URL like file:///path is a valid local path (i.e. private)
                LOGGER.debug("Ignoring private URL: %s", url)
                continue

            try:
                ips = [ip_address(parsed_url.hostname)]
            except ValueError:
                # Extra try/except here in case the DNS resolution fails, see #2348
                try:
                    ips = [
                        ip_address(ip.to_text())
                        for ip in dns.resolver.resolve(parsed_url.hostname)
                    ]
                except Exception as exc:
                    LOGGER.debug(
                        "Cannot resolve hostname %s, ignoring URL %s"
                        " (exception was: %r)",
                        parsed_url.hostname,
                        url,
                        exc,
                    )
                    continue

            private = False
            for ip in ips:
                if ip.is_private or ip.is_loopback:
                    private = True
                    break
            if private:
                LOGGER.debug("Ignoring private URL: %s", url)
                continue

        # Call the URL to get a title, if possible
        title = find_title(url)
        if not title:
            # No title found: don't handle this URL
            LOGGER.debug('No title found; ignoring URL: %s', url)
            continue

        # If the URL is over bot.config.url.shorten_url_length, shorten the URL
        tinyurl = None
        if (shorten_url_length > 0) and (len(url) > shorten_url_length):
            tinyurl = get_or_create_shorturl(bot, url)

        yield URLInfo(url, title, parsed_url.hostname, tinyurl, False)


def check_callbacks(
    bot: SopelWrapper,
    url: str,
    use_excludes: bool = True,
) -> bool:
    """Check if ``url`` is excluded or matches any URL callback patterns.

    :param bot: Sopel instance
    :param url: URL to check
    :param use_excludes: Use or ignore the configured exclusion lists
    :return: True if ``url`` is excluded or matches any URL callback pattern

    This function looks at the ``bot.memory`` for ``url_exclude`` patterns and
    it returns ``True`` if any matches the given ``url``. Otherwise, it looks
    at the ``bot``'s URL callback patterns, and it returns ``True`` if any
    matches, ``False`` otherwise.

    .. seealso::

        The :func:`~sopel.builtins.url.setup` function that defines the
        ``url_exclude`` in ``bot.memory``.

    .. versionchanged:: 7.0

        This function **does not** trigger URL callbacks anymore when ``url``
        matches a pattern.

    """
    # Check if it matches the exclusion list first
    excluded = use_excludes and any(
        regex.search(url) for regex in bot.memory["url_exclude"]
    )
    return (
        excluded or
        any(bot.search_url_callbacks(url)) or
        bot.rules.check_url_callback(bot, url)
    )


def find_title(url: str, verify: bool = True) -> Optional[str]:
    """Return the title for the given URL.

    :param verify: Whether to require a valid certificate when using https
    """
    try:
        response = requests.get(url, stream=True, verify=verify,
                                headers=DEFAULT_HEADERS)
        raw_content = b''
        for byte in response.iter_content(chunk_size=512):
            raw_content += byte
            if b'</title>' in raw_content or len(raw_content) > MAX_BYTES:
                break
        content = raw_content.decode('utf-8', errors='ignore')
        # Need to close the connection because we have not read all
        # the data
        response.close()
    except requests.exceptions.ConnectionError as e:
        LOGGER.debug("Unable to reach URL: %r: %s", url, e)
        return None
    except (
        requests.exceptions.InvalidURL,  # e.g. http:///
        UnicodeError,  # e.g. http://.example.com (urllib3<1.26)
        LocationValueError,  # e.g. http://.example.com (urllib3>=1.26)
    ):
        LOGGER.debug('Invalid URL: %s', url)
        return None

    # Some cleanup that I don't really grok, but was in the original, so
    # we'll keep it (with the compiled regexes made global) for now.
    content = TITLE_TAG_DATA.sub(r'<\1title>', content)
    content = QUOTED_TITLE.sub('', content)

    start = content.rfind('<title>')
    end = content.rfind('</title>')
    if start == -1 or end == -1:
        return None

    title = web.decode(content[start + 7:end])
    title = title.strip()[:200]

    title = ' '.join(title.split())  # cleanly remove multiple spaces

    return title or None


def get_or_create_shorturl(bot: SopelWrapper, url: str) -> Optional[str]:
    """Get or create a short URL for ``url``

    :param bot: Sopel instance
    :param url: URL to get or create a short URL for
    :return: A short URL

    It gets the short URL for ``url`` from the bot's memory if it exists.
    Otherwise, it creates a short URL (see :func:`get_tinyurl`), stores it
    into the bot's memory, then returns it.
    """
    # Check bot memory to see if the shortened URL is already in
    # memory
    if url in bot.memory['shortened_urls']:
        return bot.memory['shortened_urls'][url]

    tinyurl = get_tinyurl(url)
    bot.memory['shortened_urls'][url] = tinyurl
    return tinyurl


def get_tinyurl(url: str) -> Optional[str]:
    """Returns a shortened tinyURL link of the URL"""
    base_url = "https://tinyurl.com/api-create.php"
    tinyurl = "%s?%s" % (base_url, web.urlencode({'url': url}))
    try:
        res = requests.get(tinyurl)
        res.raise_for_status()
    except requests.exceptions.RequestException:
        return None
    # Replace text output with https instead of http to make the
    # result an HTTPS link.
    return res.text.replace("http://", "https://")



"""Sopel database module: management and tools around Sopel's datamodel.

This module defines a datamodel, using `SQLAlchemy ORM's mapping`__:

* :class:`NickIDs` and :class:`Nicknames` are used to track users
* :class:`NickValues` is used to store arbitrary values for users
* :class:`ChannelValues` is used to store arbitrary values for channels
* :class:`PluginValues` is used to store arbitrary values for plugins

These models are made available through the :class:`SopelDB` class and its
convenience methods, such as :meth:`~SopelDB.get_nick_value` or
:meth:`~SopelDB.get_channel_value`.

.. __: https://docs.sqlalchemy.org/en/14/orm/tutorial.html#declare-a-mapping
"""
from __future__ import annotations

import errno
import json
import logging
import os.path
import traceback
import typing

from sqlalchemy import Column, create_engine, ForeignKey, Integer, String
from sqlalchemy.engine.url import make_url, URL
from sqlalchemy.exc import OperationalError
from sqlalchemy.orm import declarative_base, scoped_session, sessionmaker
from sqlalchemy.sql import delete, func, select, update

from sopel.lifecycle import deprecated
from sopel.tools.identifiers import Identifier, IdentifierFactory

if typing.TYPE_CHECKING:
    from collections.abc import Iterable

    from sopel.config import Config


LOGGER = logging.getLogger(__name__)


def _deserialize(value):
    if value is None:
        return None
    # sqlite likes to return ints for strings that look like ints, even though
    # the column type is string. That's how you do dynamic typing wrong.
    value = str(value)
    # Just in case someone's mucking with the DB in a way we can't account for,
    # ignore json parsing errors
    try:
        value = json.loads(value)
    except ValueError:
        pass
    return value


BASE = declarative_base()
MYSQL_TABLE_ARGS = {'mysql_engine': 'InnoDB',
                    'mysql_charset': 'utf8mb4',
                    'mysql_collate': 'utf8mb4_unicode_ci'}


class NickIDs(BASE):
    """Nick IDs table SQLAlchemy class."""
    __tablename__ = 'nick_ids'
    nick_id = Column(Integer, primary_key=True)


class Nicknames(BASE):
    """Nicknames table SQLAlchemy class."""
    __tablename__ = 'nicknames'
    __table_args__ = MYSQL_TABLE_ARGS
    nick_id = Column(Integer, ForeignKey('nick_ids.nick_id'), primary_key=True)
    slug = Column(String(255), primary_key=True)
    canonical = Column(String(255))


class NickValues(BASE):
    """Nick values table SQLAlchemy class."""
    __tablename__ = 'nick_values'
    __table_args__ = MYSQL_TABLE_ARGS
    nick_id = Column(Integer, ForeignKey('nick_ids.nick_id'), primary_key=True)
    key = Column(String(255), primary_key=True)
    value = Column(String(255))


class ChannelValues(BASE):
    """Channel values table SQLAlchemy class."""
    __tablename__ = 'channel_values'
    __table_args__ = MYSQL_TABLE_ARGS
    channel = Column(String(255), primary_key=True)
    key = Column(String(255), primary_key=True)
    value = Column(String(255))


class PluginValues(BASE):
    """Plugin values table SQLAlchemy class."""
    __tablename__ = 'plugin_values'
    __table_args__ = MYSQL_TABLE_ARGS
    plugin = Column(String(255), primary_key=True)
    key = Column(String(255), primary_key=True)
    value = Column(String(255))


class SopelDB:
    """Database object class.

    :param config: Sopel's configuration settings
    :type config: :class:`sopel.config.Config`
    :param identifier_factory: factory for
                               :class:`~sopel.tools.identifiers.Identifier`
    :type: Callable[[:class:`str`], :class:`str`]

    This defines a simplified interface for basic, common operations on the
    bot's database. Direct access to the database is also available through
    its :attr:`engine` attribute, to serve more complex plugins' needs.

    When configured to use SQLite with a relative filename, the file is assumed
    to be in the directory named by the core setting ``homedir``.

    .. versionadded:: 5.0

    .. versionchanged:: 7.0

        Switched from direct SQLite access to :ref:`SQLAlchemy
        <sqlalchemy:overview>`, allowing users more flexibility around what type
        of database they use (especially on high-load Sopel instances, which may
        run up against SQLite's concurrent-access limitations).

    .. versionchanged:: 8.0

        An Identifier factory can be provided that will be used to instantiate
        :class:`~sopel.tools.identifiers.Identifier` when dealing with Nick or
        Channel names.

    .. seealso::

        For any advanced usage of the ORM, refer to the
        `SQLAlchemy documentation`__.

    .. __: https://docs.sqlalchemy.org/en/14/
    """

    def __init__(
        self,
        config: Config,
        identifier_factory: IdentifierFactory = Identifier,
    ) -> None:
        self.make_identifier: IdentifierFactory = identifier_factory

        if config.core.db_url is not None:
            self.url = make_url(config.core.db_url)

            # TODO: there's no way to get `config.core.db_type.choices`, but
            # it would be nice to validate this type name somehow. Shouldn't
            # affect anything, since the only thing it's ever used for is
            # checking whether the configured database is 'sqlite'.
            self.type = self.url.drivername.split('+', 1)[0]
        elif config.core.db_type == 'sqlite':
            self.type = 'sqlite'
            path = config.core.db_filename
            if path is None:
                path = os.path.join(config.core.homedir, config.basename + '.db')
            path = os.path.expanduser(path)
            if not os.path.isabs(path):
                path = os.path.normpath(os.path.join(config.core.homedir, path))
            if not os.path.isdir(os.path.dirname(path)):
                raise OSError(
                    errno.ENOENT,
                    'Cannot create database file. '
                    'No such directory: "{}". Check that configuration setting '
                    'core.db_filename is valid'.format(os.path.dirname(path)),
                    path
                )
            self.url = make_url('sqlite:///' + path)
        else:
            self.type = config.core.db_type

            query = {}
            if self.type == 'mysql':
                drivername = config.core.db_driver or 'mysql'
                query = {'charset': 'utf8mb4'}
            elif self.type == 'postgres':
                drivername = config.core.db_driver or 'postgresql'
            elif self.type == 'oracle':
                drivername = config.core.db_driver or 'oracle'
            elif self.type == 'mssql':
                drivername = config.core.db_driver or 'mssql+pymssql'
            elif self.type == 'firebird':
                drivername = config.core.db_driver or 'firebird+fdb'
            elif self.type == 'sybase':
                drivername = config.core.db_driver or 'sybase+pysybase'
            else:
                raise Exception('Unknown db_type')

            db_user = config.core.db_user  # Sometimes empty
            db_pass = config.core.db_pass  # Sometimes empty
            db_host = config.core.db_host  # Sometimes empty
            db_port = config.core.db_port  # Optional
            db_name = config.core.db_name  # Sometimes optional

            # Ensure we have all our variables defined
            if db_user is None or db_pass is None or db_host is None:
                raise Exception('Please make sure the following core '
                                'configuration values are defined: '
                                'db_user, db_pass, db_host')
            self.url = URL(drivername=drivername, username=db_user,
                           password=db_pass, host=db_host, port=db_port,
                           database=db_name, query=query)

        self.engine = create_engine(self.url, pool_recycle=3600)
        """SQLAlchemy Engine used to connect to Sopel's database.

        .. seealso::

            Read `SQLAlchemy engine`__'s documentation to know how to use it.

        .. __: https://docs.sqlalchemy.org/en/14/core/connections.html

        .. important::

            Introduced in Sopel 7, Sopel uses SQLAlchemy 1.4+. This version of
            SQLAlchemy deprecates various behaviors and methods, to prepare the
            migration to its future 2.0 version, and the new 2.x style.

            Sopel doesn't enforce the new 2.x style yet. This will be modified
            in Sopel 9 by using the ``future=True`` flag on the engine.

            You can read more about the `migration guide from 1.x to 2.x`__, as
            Sopel will ensure in a future version that it is compatible with
            the new style.

        .. __: https://docs.sqlalchemy.org/en/14/changelog/migration_20.html
        """

        # Catch any errors connecting to database
        try:
            self.engine.connect()
        except OperationalError:
            print("OperationalError: Unable to connect to database.")
            raise

        # Create our tables
        BASE.metadata.create_all(self.engine)

        self.ssession = scoped_session(
            sessionmaker(bind=self.engine, future=True))

    def connect(self):
        """Get a direct database connection.

        :return: a proxied DBAPI connection object; see
                 :meth:`sqlalchemy.engine.Engine.raw_connection()`

        .. important::

           The :attr:`~sopel.config.core_section.CoreSection.db_type` in use
           can change how the raw connection object behaves. You probably want
           to use :meth:`session` and the SQLAlchemy ORM in new plugins, and
           officially support only Sopel 7.0+.

           Note that :meth:`session` is not available in Sopel versions prior
           to 7.0. If your plugin needs to be compatible with older Sopel
           releases, your code *should* use SQLAlchemy via :meth:`session` if
           it is available (Sopel 7.0+) and fall back to direct SQLite access
           via :meth:`connect` if it is not (Sopel 6.x).

           We discourage *publishing* plugins that don't work with all
           supported databases, but you're obviously welcome to take shortcuts
           and support only the engine(s) you need in *private* plugins.

        """
        if self.type != 'sqlite':
            # log non-sqlite uses of raw connections for troubleshooting, since
            # unless the developer had a good reason to use this instead of
            # `session()`, it indicates the plugin was written before Sopel 7.0
            # and might not work right when connected to non-sqlite DBs
            LOGGER.info(
                "Raw connection requested when 'db_type' is not 'sqlite':\n"
                "Consider using 'db.session()' to get a SQLAlchemy session "
                "instead here:\n%s",
                traceback.format_list(traceback.extract_stack()[:-1])[-1][:-1])
        return self.engine.raw_connection()

    def session(self):
        """Get a SQLAlchemy Session object.

        :rtype: :class:`sqlalchemy.orm.session.Session`

        .. versionadded:: 7.0

        .. note::

           If your plugin needs to remain compatible with Sopel versions prior
           to 7.0, you can use :meth:`connect` to get a raw connection. See
           its documentation for relevant warnings and compatibility caveats.

        """
        return self.ssession()

    @deprecated('Use SopelDB.engine directly', version='8.0', removed_in='9.0')
    def execute(self, *args, **kwargs):
        """Execute an arbitrary SQL query against the database.

        :return: the query results
        :rtype: :class:`sqlalchemy.engine.Result`

        The ``Result`` object returned is a wrapper around a ``Cursor`` object
        as specified by :pep:`249`.

        .. deprecated:: 8.0

            This method will be removed in Sopel 9, following the deprecation
            of SQLAlchemy's :meth:`sqlalchemy.engine.Engine.execute`.

            To perform a raw SQL query, use the :class:`~SopelDB.engine`
            attribute as per the migration guide from SQLAlchemy::

                from sqlalchemy.sql import text

                def my_command(bot, trigger):
                    raw_sql = ' ... '  # your raw SQL
                    # get a connection as a context manager
                    with bot.db.engine.connect() as conn:
                        res = conn.execute(text(raw_sql))
                        data = res.fetchall()

                    # do something with your data here

        .. seealso::

            Read the `migration guide from 1.x style to 2.x style`__ by
            SQLAlchemy to learn more about using SQLALchemy's engine and
            connection.

        .. __: https://docs.sqlalchemy.org/en/14/changelog/migration_20.html
        """
        return self.engine.execute(*args, **kwargs)

    def get_uri(self) -> URL:
        """Return a direct URL for the database.

        :return: the database connection URI
        :rtype: str

        This can be used to connect from a plugin using another SQLAlchemy
        instance, for example, without sharing the bot's connection.
        """
        return self.url

    # NICK FUNCTIONS

    def get_nick_id(self, nick: str, create: bool = False) -> int:
        """Return the internal identifier for a given nick.

        :param nick: the nickname for which to fetch an ID
        :param create: whether to create an ID if one does not exist
                       (set to ``False`` by default)
        :raise ValueError: if no ID exists for the given ``nick`` and ``create``
                           is set to ``False``
        :raise ~sqlalchemy.exc.SQLAlchemyError: if there is a database error

        The nick ID is shared across all of a user's aliases, assuming their
        nicks have been grouped together.

        .. versionchanged:: 8.0

            The ``create`` parameter is now ``False`` by default.

        .. seealso::

            Alias/group management functions: :meth:`alias_nick`,
            :meth:`unalias_nick`, :meth:`merge_nick_groups`, and
            :meth:`forget_nick_group`.

        """
        slug = self.make_identifier(nick).lower()
        with self.session() as session:
            nickname = session.execute(
                select(Nicknames).where(Nicknames.slug == slug)
            ).scalar_one_or_none()

            if nickname is None:
                # see if it needs case-mapping migration
                nickname = session.execute(
                    select(Nicknames)
                    .where(Nicknames.slug == Identifier._lower_swapped(nick))
                ).scalar_one_or_none()

                if nickname is not None:
                    # it does!
                    nickname.slug = slug
                    session.commit()

            if nickname is None:  # "is /* still */ None", if Python had inline comments
                if not create:
                    raise ValueError('No ID exists for the given nick')
                # Generate a new ID
                nick_id = NickIDs()
                session.add(nick_id)
                session.commit()

                # Create a new Nickname
                nickname = Nicknames(
                    nick_id=nick_id.nick_id,
                    slug=slug,
                    canonical=nick,
                )
                session.add(nickname)
                session.commit()
            return nickname.nick_id

    def alias_nick(self, nick: str, alias: str) -> None:
        """Create an alias for a nick.

        :param nick: an existing nickname
        :param alias: an alias by which ``nick`` should also be known
        :raise ValueError: if the ``alias`` already exists
        :raise ~sqlalchemy.exc.SQLAlchemyError: if there is a database error

        .. seealso::

            To merge two *existing* nick groups, use :meth:`merge_nick_groups`.

            To remove an alias created with this function, use
            :meth:`unalias_nick`.

        """
        slug = self.make_identifier(alias).lower()
        nick_id = self.get_nick_id(nick, create=True)
        with self.session() as session:
            result = session.execute(
                select(Nicknames)
                .where(Nicknames.slug == slug)
                .where(Nicknames.canonical == alias)
            ).scalar_one_or_none()
            if result:
                raise ValueError('Alias already exists.')
            nickname = Nicknames(
                nick_id=nick_id,
                slug=slug,
                canonical=alias,
            )
            session.add(nickname)
            session.commit()

    def set_nick_value(self, nick: str, key: str, value: typing.Any) -> None:
        """Set or update a value in the key-value store for ``nick``.

        :param nick: the nickname with which to associate the ``value``
        :param key: the name by which this ``value`` may be accessed later
        :param value: the value to set for this ``key`` under ``nick``
        :raise ~sqlalchemy.exc.SQLAlchemyError: if there is a database error

        The ``value`` can be any of a range of types; it need not be a string.
        It will be serialized to JSON before being stored and decoded
        transparently upon retrieval.

        .. seealso::

            To retrieve a value set with this method, use
            :meth:`get_nick_value`.

            To delete a value set with this method, use
            :meth:`delete_nick_value`.

        """
        value = json.dumps(value, ensure_ascii=False)
        nick_id = self.get_nick_id(nick, create=True)
        with self.session() as session:
            result = session.execute(
                select(NickValues)
                .where(NickValues.nick_id == nick_id)
                .where(NickValues.key == key)
            ).scalar_one_or_none()

            # NickValue exists, update
            if result:
                result.value = value
                session.commit()
            # DNE - Insert
            else:
                new_nickvalue = NickValues(
                    nick_id=nick_id,
                    key=key,
                    value=value,
                )
                session.add(new_nickvalue)
                session.commit()

    def delete_nick_value(self, nick: str, key: str) -> None:
        """Delete a value from the key-value store for ``nick``.

        :param nick: the nickname whose values to modify
        :param key: the name of the value to delete
        :raise ~sqlalchemy.exc.SQLAlchemyError: if there is a database error

        .. seealso::

            To set a value in the first place, use :meth:`set_nick_value`.

            To retrieve a value instead of deleting it, use
            :meth:`get_nick_value`.

        """
        try:
            nick_id = self.get_nick_id(nick)
        except ValueError:
            # there's nothing to do if the nick doesn't exist
            return

        with self.session() as session:
            result = session.execute(
                select(NickValues)
                .where(NickValues.nick_id == nick_id)
                .where(NickValues.key == key)
            ).scalar_one_or_none()
            # NickValue exists, delete
            if result:
                session.delete(result)
                session.commit()

    def get_nick_value(
        self,
        nick: str,
        key: str,
        default: typing.Optional[typing.Any] = None
    ) -> typing.Optional[typing.Any]:
        """Get a value from the key-value store for ``nick``.

        :param nick: the nickname whose values to access
        :param key: the name by which the desired value was saved
        :param default: value to return if ``key`` does not have a value set
                        (optional)
        :raise ~sqlalchemy.exc.SQLAlchemyError: if there is a database error

        .. versionadded:: 7.0

            The ``default`` parameter.

        .. seealso::

            To set a value for later retrieval with this method, use
            :meth:`set_nick_value`.

            To delete a value instead of retrieving it, use
            :meth:`delete_nick_value`.

        """
        slug = self.make_identifier(nick).lower()
        with self.session() as session:
            result = session.execute(
                select(NickValues)
                .where(Nicknames.nick_id == NickValues.nick_id)
                .where(Nicknames.slug == slug)
                .where(NickValues.key == key)
            ).scalar_one_or_none()

            if result is not None:
                result = result.value
            elif default is not None:
                result = default

            return _deserialize(result)

    def unalias_nick(self, alias: str) -> None:
        """Remove an alias.

        :param alias: an alias with at least one other nick in its group
        :raise ValueError: if there is not at least one other nick in the
                           group, or the ``alias`` is not known
        :raise ~sqlalchemy.exc.SQLAlchemyError: if there is a database error

        .. seealso::

            To delete an entire group, use :meth:`forget_nick_group`.

            To *add* an alias for a nick, use :meth:`alias_nick`.

        """
        slug = self.make_identifier(alias).lower()
        nick_id = self.get_nick_id(alias)
        with self.session() as session:
            count = session.scalar(
                select(func.count()).select_from(Nicknames)
                .where(Nicknames.nick_id == nick_id)
            )
            if count <= 1:
                raise ValueError('Given alias is the only entry in its group.')
            session.execute(
                delete(Nicknames)
                .where(Nicknames.slug == slug)
                .execution_options(synchronize_session="fetch")
            )
            session.commit()

    def forget_nick_group(self, nick: str) -> None:
        """Remove a nickname, all of its aliases, and all of its stored values.

        :param nick: one of the nicknames in the group to be deleted
        :raise ValueError: if the ``nick`` does not exist in the database
        :raise ~sqlalchemy.exc.SQLAlchemyError: if there is a database error

        .. important::

            This is otherwise known as The Nuclear Option. Be *very* sure that
            you want to do this.

        """
        nick_id = self.get_nick_id(nick)
        with self.session() as session:
            session.execute(
                delete(Nicknames)
                .where(Nicknames.nick_id == nick_id)
                .execution_options(synchronize_session="fetch")
            )
            session.execute(
                delete(NickValues)
                .where(NickValues.nick_id == nick_id)
                .execution_options(synchronize_session="fetch")
            )
            session.commit()

    @deprecated(
        version='8.0',
        removed_in='9.0',
        reason="Renamed to `forget_nick_group`",
    )
    def delete_nick_group(self, nick: str) -> None:  # pragma: nocover
        self.forget_nick_group(nick)

    def merge_nick_groups(self, first_nick: str, second_nick: str) -> None:
        """Merge two nick groups.

        :param first_nick: one nick in the first group to merge
        :param second_nick: one nick in the second group to merge
        :raise ~sqlalchemy.exc.SQLAlchemyError: if there is a database error

        Takes two nicks, which may or may not be registered. Unregistered nicks
        will be registered. Keys which are set for only one of the given nicks
        will be preserved. Where both nicks have values for a given key, the
        value set for the ``first_nick`` will be used.

        A nick group can contain one or many nicknames. Groups containing more
        than one nickname can be created with this function, or by using
        :meth:`alias_nick` to add aliases.

        Note that merging of data only applies to the native key-value store.
        Plugins which define their own tables relying on the nick table will
        need to handle their own merging separately.
        """
        first_id = self.get_nick_id(first_nick, create=True)
        second_id = self.get_nick_id(second_nick, create=True)
        with self.session() as session:
            # Get second_id's values
            results = session.execute(
                select(NickValues).where(NickValues.nick_id == second_id)
            ).scalars()

            # Update first_id with second_id values if first_id doesn't have that key
            for row in results:
                first_res = session.execute(
                    select(NickValues)
                    .where(NickValues.nick_id == first_id)
                    .where(NickValues.key == row.key)
                ).scalar_one_or_none()

                if not first_res:
                    self.set_nick_value(
                        first_nick, row.key, _deserialize(row.value))

            session.execute(
                delete(NickValues)
                .where(NickValues.nick_id == second_id)
                .execution_options(synchronize_session="fetch")
            )
            session.execute(
                update(Nicknames)
                .where(Nicknames.nick_id == second_id)
                .values(nick_id=first_id)
                .execution_options(synchronize_session="fetch")
            )
            session.commit()

    # CHANNEL FUNCTIONS

    def get_channel_slug(self, chan: str) -> str:
        """Return the case-normalized representation of ``channel``.

        :param channel: the channel name to normalize, with prefix (required)
        :return: the case-normalized channel name (or "slug" representation)

        This is useful to make sure that a channel name is stored consistently
        in both the bot's own database and third-party plugins'
        databases/files, without regard for variation in case between
        different clients and/or servers on the network.
        """
        slug = self.make_identifier(chan).lower()

        with self.session() as session:
            # Always migrate from old casemapping
            session.execute(
                update(ChannelValues)
                .where(ChannelValues.channel == Identifier._lower_swapped(chan))
                .values(channel=slug)
                .execution_options(synchronize_session="fetch")
            )
            session.commit()

        return slug

    def set_channel_value(
        self,
        channel: str,
        key: str,
        value: typing.Any,
    ) -> None:
        """Set or update a value in the key-value store for ``channel``.

        :param channel: the channel with which to associate the ``value``
        :param key: the name by which this ``value`` may be accessed later
        :param value: the value to set for this ``key`` under ``channel``
        :raise ~sqlalchemy.exc.SQLAlchemyError: if there is a database error

        The ``value`` can be any of a range of types; it need not be a string.
        It will be serialized to JSON before being stored and decoded
        transparently upon retrieval.

        .. seealso::

            To retrieve a value set with this method, use
            :meth:`get_channel_value`.

            To delete a value set with this method, use
            :meth:`delete_channel_value`.

        """
        channel = self.get_channel_slug(channel)
        value = json.dumps(value, ensure_ascii=False)
        with self.session() as session:
            result = session.execute(
                select(ChannelValues)
                .where(ChannelValues.channel == channel)
                .where(ChannelValues.key == key)
            ).scalar_one_or_none()

            # ChannelValue exists, update
            if result:
                result.value = value
                session.commit()
            # DNE - Insert
            else:
                new_channelvalue = ChannelValues(
                    channel=channel,
                    key=key,
                    value=value,
                )
                session.add(new_channelvalue)
                session.commit()

    def delete_channel_value(self, channel: str, key: str) -> None:
        """Delete a value from the key-value store for ``channel``.

        :param channel: the channel whose values to modify
        :param key: the name of the value to delete
        :raise ~sqlalchemy.exc.SQLAlchemyError: if there is a database error

        .. seealso::

            To set a value in the first place, use :meth:`set_channel_value`.

            To retrieve a value instead of deleting it, use
            :meth:`get_channel_value`.

        """
        channel = self.get_channel_slug(channel)
        with self.session() as session:
            session.execute(
                delete(ChannelValues)
                .where(
                    ChannelValues.channel == channel,
                    ChannelValues.key == key
                ).execution_options(synchronize_session="fetch")
            )
            session.commit()

    def get_channel_value(
        self,
        channel: str,
        key: str,
        default: typing.Optional[typing.Any] = None,
    ) -> typing.Any:
        """Get a value from the key-value store for ``channel``.

        :param channel: the channel whose values to access
        :param key: the name by which the desired value was saved
        :param default: value to return if ``key`` does not have a value set
                        (optional)
        :raise ~sqlalchemy.exc.SQLAlchemyError: if there is a database error

        .. versionadded:: 7.0

            The ``default`` parameter.

        .. seealso::

            To set a value for later retrieval with this method, use
            :meth:`set_channel_value`.

            To delete a value instead of retrieving it, use
            :meth:`delete_channel_value`.

        """
        channel = self.get_channel_slug(channel)
        with self.session() as session:
            result = session.execute(
                select(ChannelValues)
                .where(ChannelValues.channel == channel)
                .where(ChannelValues.key == key)
            ).scalar_one_or_none()
            if result is not None:
                result = result.value
            elif default is not None:
                result = default
            return _deserialize(result)

    def forget_channel(self, channel: str) -> None:
        """Remove all of a channel's stored values.

        :param channel: the name of the channel for which to delete values
        :raise ~sqlalchemy.exc.SQLAlchemyError: if there is a database error

        .. important::

            This is a Nuclear Option. Be *very* sure that you want to do it.

        """
        channel = self.get_channel_slug(channel)
        with self.session() as session:
            session.execute(
                delete(ChannelValues)
                .where(ChannelValues.channel == channel)
            )
            session.commit()

    # PLUGIN FUNCTIONS

    def set_plugin_value(
        self,
        plugin: str,
        key: str,
        value: typing.Any,
    ) -> None:
        """Set or update a value in the key-value store for ``plugin``.

        :param plugin: the plugin name with which to associate the ``value``
        :param key: the name by which this ``value`` may be accessed later
        :param value: the value to set for this ``key`` under ``plugin``
        :raise ~sqlalchemy.exc.SQLAlchemyError: if there is a database error

        The ``value`` can be any of a range of types; it need not be a string.
        It will be serialized to JSON before being stored and decoded
        transparently upon retrieval.

        .. seealso::

            To retrieve a value set with this method, use
            :meth:`get_plugin_value`.

            To delete a value set with this method, use
            :meth:`delete_plugin_value`.

        """
        plugin = plugin.lower()
        value = json.dumps(value, ensure_ascii=False)
        with self.session() as session:
            result = session.execute(
                select(PluginValues)
                .where(PluginValues.plugin == plugin)
                .where(PluginValues.key == key)
            ).scalar_one_or_none()
            # PluginValue exists, update
            if result:
                result.value = value
                session.commit()
            # DNE - Insert
            else:
                new_pluginvalue = PluginValues(plugin=plugin, key=key, value=value)
                session.add(new_pluginvalue)
                session.commit()

    def delete_plugin_value(self, plugin: str, key: str) -> None:
        """Delete a value from the key-value store for ``plugin``.

        :param plugin: the plugin name whose values to modify
        :param key: the name of the value to delete
        :raise ~sqlalchemy.exc.SQLAlchemyError: if there is a database error

        .. seealso::

            To set a value in the first place, use :meth:`set_plugin_value`.

            To retrieve a value instead of deleting it, use
            :meth:`get_plugin_value`.

        """
        plugin = plugin.lower()
        with self.session() as session:
            result = session.execute(
                select(PluginValues)
                .where(PluginValues.plugin == plugin)
                .where(PluginValues.key == key)
            ).scalar_one_or_none()
            # PluginValue exists, update
            if result:
                session.delete(result)
                session.commit()

    def get_plugin_value(
        self,
        plugin: str,
        key: str,
        default: typing.Optional[typing.Any] = None,
    ) -> typing.Optional[typing.Any]:
        """Get a value from the key-value store for ``plugin``.

        :param plugin: the plugin name whose values to access
        :param key: the name by which the desired value was saved
        :param default: value to return if ``key`` does not have a value set
                        (optional)
        :raise ~sqlalchemy.exc.SQLAlchemyError: if there is a database error

        .. versionadded:: 7.0

            The ``default`` parameter.

        .. seealso::

            To set a value for later retrieval with this method, use
            :meth:`set_plugin_value`.

            To delete a value instead of retrieving it, use
            :meth:`delete_plugin_value`.

        """
        plugin = plugin.lower()
        with self.session() as session:
            result = session.execute(
                select(PluginValues)
                .where(PluginValues.plugin == plugin)
                .where(PluginValues.key == key)
            ).scalar_one_or_none()

            if result is not None:
                result = result.value
            elif default is not None:
                result = default
            return _deserialize(result)

    def forget_plugin(self, plugin: str) -> None:
        """Remove all of a plugin's stored values.

        :param plugin: the name of the plugin for which to delete values
        :raise ~sqlalchemy.exc.SQLAlchemyError: if there is a database error

        .. important::

            This is a Nuclear Option. Be *very* sure that you want to do it.

        """
        plugin = plugin.lower()
        with self.session() as session:
            session.execute(
                delete(PluginValues).where(PluginValues.plugin == plugin)
            )
            session.commit()

    # NICK AND CHANNEL FUNCTIONS

    def get_nick_or_channel_value(
        self,
        name: str,
        key: str,
        default: typing.Any | None = None
    ) -> typing.Any | None:
        """Get a value from the key-value store for ``name``.

        :param name: nick or channel whose values to access
        :param key: the name by which the desired value was saved
        :param default: value to return if ``key`` does not have a value set
                        (optional)
        :raise ~sqlalchemy.exc.SQLAlchemyError: if there is a database error

        .. versionadded:: 7.0

            The ``default`` parameter.

        This is useful for common logic that is shared between both users and
        channels, as it will fetch the appropriate value based on what type of
        ``name`` it is given.

        .. seealso::

            To get a value for a nick specifically, use :meth:`get_nick_value`.

            To get a value for a channel specifically, use
            :meth:`get_channel_value`.

        """
        if not isinstance(name, Identifier):
            identifier = self.make_identifier(name)
        else:
            identifier = typing.cast('Identifier', name)

        if identifier.is_nick():
            return self.get_nick_value(identifier, key, default)
        else:
            return self.get_channel_value(identifier, key, default)

    def get_preferred_value(
        self,
        names: Iterable[str],
        key: str,
    ) -> typing.Optional[typing.Any]:
        """Get a value for the first name which has it set.

        :param names: a list of channel names and/or nicknames
        :param key: the name by which the desired value was saved
        :return: the value for ``key`` from the first ``name`` which has it set,
                 or ``None`` if none of the ``names`` has it set
        :raise ~sqlalchemy.exc.SQLAlchemyError: if there is a database error

        This is useful for logic that needs to customize its output based on
        settings stored in the database. For example, it can be used to fall
        back from the triggering user's setting to the current channel's setting
        in case the user has not configured their setting.

        .. note::

            This is the only ``get_*_value()`` method that does not support
            passing a ``default``. Try to avoid using it on ``key``\\s which
            might have ``None`` as a valid value, to avoid ambiguous logic.

        """
        for name in names:
            value = self.get_nick_or_channel_value(name, key)
            if value is not None:
                return value

        # Explicit return for type check
        return None



"""Management of IRC capability negotiation.

.. versionadded:: 8.0

This module contains the :class:`Capabilities` class which tracks the state of
the capability negotiation with the IRC server: it can store and update the
state of available and enabled capabilities.

.. important::

    Plugin authors should not instantiate this class directly, as the bot
    exposes an instance through its :attr:`~sopel.irc.AbstractBot.capabilities`
    attribute.

    All state handling methods (such as :meth:`Capabilities.handle_ls`) are
    used by the ``coretasks`` plugin, and should not be used outside.

"""
from __future__ import annotations

from typing import (
    NamedTuple,
    Optional,
    TYPE_CHECKING,
)


if TYPE_CHECKING:
    from sopel.bot import SopelWrapper
    from sopel.trigger import Trigger


class CapabilityInfo(NamedTuple):
    """Capability metadata.

    This contains the details of a capability: if :attr:`available`,
    :attr:`enabled`, and its :attr:`params` (if advertised).

    .. note::

        You can get a capability's info through
        :meth:`Capabilities.get_capability_info`.

    """
    name: str
    """Name of the capability.

    The name of a capability is the name as it appears in the ``CAP LS``
    subcommand, such as ``multi-prefix`` or ``sasl``.
    """
    params: Optional[str]
    """Advertised parameters for this capability.

    When a server supports ``CAP`` version 302, capabilities can have
    parameters. The format and the meaning of the parameters depend on the
    capability itself.

    For example, the ``sasl`` capability can provide the list of SASL
    mechanisms the server supports, such as ``PLAIN,EXTERNAL``.
    """
    available: bool
    """Flag to tell if the server advertises this capability or not.

    This is ``True`` if the ``CAP LS`` subcommand contains the capability.
    """
    enabled: bool
    """Flag to tell if the capability is enabled on the server."""


class Capabilities:
    """Capabilities negotiated with the server.

    This stores a representation of the capability negotiation state between
    the bot and the server: it stores the list of :attr:`available` and
    :attr:`enabled` capabilities, and can track the state by handling various
    ``CAP`` subcommands:

    * :meth:`handle_ls` for ``CAP LS``
    * :meth:`handle_ack` for ``CAP ACK``
    * :meth:`handle_nak` for ``CAP NAK``
    * :meth:`handle_new` for ``CAP NEW``
    * :meth:`handle_del` for ``CAP ADD``
    """
    def __init__(self) -> None:
        self._available: dict[str, Optional[str]] = {}
        self._enabled: set[str] = set()

    def get_capability_info(self, name: str) -> CapabilityInfo:
        """Retrieve metadata about a capability.

        The returned :class:`CapabilityInfo` will tell if the capability is
        advertised by the server; and if so, its parameters and whether it is
        enabled.

        If the capability is unknown, then its ``available`` attribute will be
        ``False``, and its ``params`` attribute will be ``None``.
        """
        params = self._available.get(name)
        return CapabilityInfo(
            name,
            params,
            name in self._available,
            name in self._enabled,
        )

    @property
    def available(self) -> dict[str, Optional[str]]:
        """Dict of available server capabilities.

        Each key is the name of a capability advertised by the server, and each
        value is the parameters as advertised (if any) for this capability.

        If a capability is not in this ``dict``, it means the server doesn't
        advertise it, and it cannot be requested.
        """
        return dict(self._available.items())  # return a copy

    @property
    def enabled(self) -> frozenset[str]:
        """Set of enabled server capabilities.

        Each element is the name of a capability that is enabled on the server.
        """
        return frozenset(self._enabled)

    def is_available(self, name: str) -> bool:
        """Tell if the capability ``name`` is available on the server."""
        return name in self._available

    def is_enabled(self, name: str) -> bool:
        """Tell if the capability ``name`` is enabled on the server."""
        return name in self._enabled

    def handle_ls(self, bot: SopelWrapper, trigger: Trigger) -> bool:
        """Handle a ``CAP LS`` command.

        This method behaves as a plugin callable with its ``bot`` and
        ``trigger`` arguments, with the precise goals to handle ``CAP LS``
        command only, registering available capabilities.

        Then it returns if there is no more ``LS`` command to handle (in case
        of multi-line ``LS``).
        """
        # dev checking
        assert trigger.event == 'CAP'
        assert trigger.args[1] == 'LS'

        # extracting capabilities
        for available_capability in trigger.split():
            name, *params = available_capability.split('=', maxsplit=1)
            self._available[name] = params[0] if params else None

        # return if multiline or not
        multiline = trigger.args[2] == '*'
        return not multiline

    def handle_ack(
        self,
        bot: SopelWrapper,
        trigger: Trigger,
    ) -> tuple[str, ...]:
        """Handle a ``CAP ACK`` command.

        This method behaves as a plugin callable with its ``bot`` and
        ``trigger`` arguments, with the precise goals to handle ``CAP ACK``
        command only, registering enabled or disabled capabilities and tracking
        acknowledgement.

        Then it returns the list of acknowledged capability requests.
        """
        # dev checking
        assert trigger.event == 'CAP'
        assert trigger.args[1] == 'ACK'

        # extracting capabilities
        ack_capabilities = tuple(sorted(trigger.split()))

        # update enabled capabilities
        for ack_capability in ack_capabilities:
            if ack_capability.startswith('-'):
                # ACK a disable request
                cap_reversed = ack_capability.lstrip('-')
                if cap_reversed in self._enabled:
                    self._enabled.remove(cap_reversed)
            else:
                # ACK an enable request
                self._enabled.add(ack_capability)

        # return set of acknowledged capabilities
        return ack_capabilities

    def handle_nak(
        self,
        bot: SopelWrapper,
        trigger: Trigger,
    ) -> tuple[str, ...]:
        """Handle a ``CAP NAK`` command.

        This method behaves as a plugin callable with its ``bot`` and
        ``trigger`` arguments, with the precise goals to handle ``CAP NAK``
        command only.

        Then it returns the list of denied capability requests.
        """
        # dev checking
        assert trigger.event == 'CAP'
        assert trigger.args[1] == 'NAK'

        # extracting capabilities
        nack_capabilities = tuple(sorted((trigger.split())))

        # return tuple of not acknowledged capabilities
        return nack_capabilities

    def handle_new(
        self,
        bot: SopelWrapper,
        trigger: Trigger,
    ) -> tuple[str, ...]:
        """Handle a ``CAP NEW`` command.

        This method behaves as a plugin callable with its ``bot`` and
        ``trigger`` arguments, with the precise goals to handle ``CAP NEW``
        command only.

        It registers which new capabilities are available, then returns this
        list.
        """
        # dev checking
        assert trigger.event == 'CAP'
        assert trigger.args[1] == 'NEW'

        # extracting capabilities
        cap_new: set[str] = set()
        for available_capability in trigger.split():
            name, *params = available_capability.split('=', maxsplit=1)
            self._available[name] = params[0] if params else None
            cap_new.add(name)

        return tuple(sorted(cap_new))

    def handle_del(
        self,
        bot: SopelWrapper,
        trigger: Trigger,
    ) -> tuple[str, ...]:
        """Handle a ``CAP DEL`` command.

        This method behaves as a plugin callable with its ``bot`` and
        ``trigger`` arguments, with the precise goals to handle ``CAP DEL``
        command only.

        It registers which capabilities are not available and not enabled
        anymore, then returns this list.
        """
        # dev checking
        assert trigger.event == 'CAP'
        assert trigger.args[1] == 'DEL'

        # extracting capabilities
        cap_del: set[str] = set()
        for name in trigger.split():
            if name in self._available:
                del self._available[name]
            if name in self._enabled:
                self._enabled.remove(name)

            cap_del.add(name)

        return tuple(sorted(cap_del))



"""Mode management for IRC channels.

The :class:`ModeParser` class is used internally by the bot to parse ``MODE``
messages for channels. User modes are not parsed (yet), as the bot doesn't
manage them.

The goal of the parser is to return a :class:`ModeMessage` containing the
actions represented by the raw message:

* channel modes added/removed (including their parameters, if any)
* privileges added/removed for user(s) in a channel

Errors (ignored modes and unused parameters) are also included, mostly for
detecting when an IRC server is not conforming to specifications.

.. important::

    This is mostly for internal use only as plugin developers should be more
    interested in :attr:`privileges<sopel.tools.target.Channel.privileges>`
    rather than how Sopel knows about them.

    The interface of this module is subject to change between Sopel releases
    without advance notice, even in patch versions.

.. seealso::

    https://modern.ircdocs.horse/#mode-message

.. versionadded:: 8.0
"""
from __future__ import annotations

import enum
import logging
from typing import (
    NamedTuple,
    Optional,
    Tuple,
    TYPE_CHECKING,
)

if TYPE_CHECKING:
    from collections.abc import Iterator


ModeTuple = Tuple[str, bool]
"""Tuple of mode information: ``(mode, is_added)``.

Where ``mode`` is the mode or privilege letter and ``is_added`` tells if
the mode or privilege wants to be added or removed.

This type alias represents the basic information for each mode found when
parsing a modestring like ``+abc-efg``. In that example mode ``a`` and mode
``f`` would be represented as these tuples: ``('a', True)`` and
``('f', False)``.
"""
ModeDetails = Tuple[str, str, bool, Optional[str]]
"""Tuple of mode details as ``(letter, mode, is_added, param)``.

Where ``type`` is the mode type (such as A, B, C, D); ``mode`` is the mode
letter; ``is_added`` tells if the mode should be added or removed; and
``param`` is an optional parameter value for that mode only when necessary.
"""
PrivilegeDetails = Tuple[str, bool, str]
"""Tuple of privilege details as ``(mode, is_added, param)``

Where ``privilege`` is the privilege letter; ``is_added`` tells if the
privilege should be added or removed; and ``target`` is the target for that
privilege.
"""


LOGGER = logging.getLogger(__name__)


class ParamRequired(enum.Enum):
    """Enum of param requirement for mode types."""
    ALWAYS = 'always'
    """The mode type always requires a parameter."""
    ADDED = 'added'
    """The mode type requires a parameter only when the mode is added."""
    REMOVED = 'removed'
    """The mode type requires a parameter only when the mode is removed."""
    NEVER = 'never'
    """The mode type never requires a parameter."""


DEFAULT_MODETYPE_PARAM_CONFIG = {
    'A': ParamRequired.ALWAYS,
    'B': ParamRequired.ALWAYS,
    'C': ParamRequired.ADDED,
    'D': ParamRequired.NEVER,
}
"""Default parameter requirements for mode types."""


class ModeException(Exception):
    """Base exception class for mode management."""


class ModeTypeUnknown(ModeException):
    """Exception when a mode's type is unknown or cannot be determined."""
    def __init__(self, mode: str) -> None:
        super().__init__('Unknown type for mode %s' % mode)


class ModeTypeImproperlyConfigured(ModeException):
    """Exception when the mode's type management is not configured properly."""
    def __init__(self, mode: str, letter: str) -> None:
        message = 'Type {mode} for mode {letter} is not properly configured.'
        super().__init__(message.format(mode=mode, letter=letter))


def parse_modestring(modestring: str) -> Iterator[ModeTuple]:
    """Parse a modestring like ``+abc-def`` and yield :class:`ModeTuple`."""
    is_added = True
    for char in modestring:
        if char in '+-':
            is_added = char == '+'
            continue
        yield (char, is_added)


class ModeMessage(NamedTuple):
    """Mode message with channel's modes and channel's privileges."""
    modes: tuple[ModeDetails, ...]
    """Tuple of added and removed modes.

    Each item is a :class:`ModeDetails`.
    """
    privileges: tuple[PrivilegeDetails, ...]
    """Tuple of added and removed privileges.

    Each item is a :class:`PrivilegeDetails`.
    """
    ignored_modes: tuple[ModeTuple, ...]
    """Ignored modes when they are unknown or there is a missing parameter.

    Each item is a :class:`ModeTuple`.
    """
    leftover_params: tuple[str, ...]
    """Parameters not used by any valid mode or privilege."""


class ModeParser:
    """ModeMessage parser for IRC's ``MODE`` messages for channel modes."""
    PRIVILEGES: set[str] = {
        "v",  # VOICE
        "h",  # HALFOP
        "o",  # OP
        "a",  # ADMIN
        "q",  # OWNER
        "y",  # OPER
        "Y",  # OPER
    }
    """Set of user privileges used by default."""

    CHANMODES = {
        'A': tuple('beI'),
        'B': tuple('k'),
        'C': tuple('l'),
        'D': tuple('Oimnpsrt'),
    }
    """Default CHANMODES per :rfc:`2811`.

    .. note::

        Mode ``a`` has been removed from the default list, as it appears
        to be a relic of the past and is more commonly used as a privilege.

        Mode ``q`` has been removed too, as it is commonly used as a privilege.

        If a server is unhappy with these defaults, they should advertise
        ``CHANMODES`` and ``PREFIX`` properly.

    """

    def __init__(
        self,
        chanmodes: dict[str, tuple[str, ...]] = CHANMODES,
        type_params: dict[str, ParamRequired] = DEFAULT_MODETYPE_PARAM_CONFIG,
        privileges: set[str] = PRIVILEGES,
    ) -> None:
        self.chanmodes: dict[str, tuple[str, ...]] = dict(chanmodes)
        """Map of mode types (``str``) to their lists of modes (``tuple``).

        This map should come from ``ISUPPORT``, usually through
        :attr:`bot.isupport.CHANMODES <sopel.irc.isupport.ISupport.CHANMODES>`.
        """
        self.type_params = dict(type_params)
        """Map of mode types (``str``) with their param requirements.

        This map defaults to :data:`DEFAULT_MODETYPE_PARAM_CONFIG`.
        """
        self.privileges = set(privileges)
        """Set of valid user privileges.

        This set should come from ``ISUPPORT``, usually through
        :attr:`bot.isupport.PREFIX <sopel.irc.isupport.ISupport.PREFIX>`.

        If a server doesn't advertise its prefixes for user privileges,
        :attr:`PRIVILEGES` will be used as a default value.
        """

    def get_mode_type(self, mode: str) -> str:
        """Retrieve the type of ``mode``.

        :raise ModeTypeUnknown: if the mode's type cannot be determined
        :return: the mode's type as defined by :attr:`chanmodes`

        ::

            >>> mm = ModeParser({'A': tuple('beI'), 'B': tuple('k')}, {})
            >>> mm.get_mode_type('b')
            'A'
            >>> mm.get_mode_type('k')
            'B'

        This method will raise a :exc:`ModeTypeUnknown` if the mode is unknown,
        including the case where ``mode`` is actually a user privilege such as
        ``v``.
        """
        for letter, modes in self.chanmodes.items():
            if mode in modes:
                return letter
        raise ModeTypeUnknown(mode)

    def get_mode_info(self, mode: str, is_added: bool) -> tuple[str, bool]:
        """Retrieve ``mode``'s information when added or removed.

        :raise ModeTypeUnknown: when the mode's type is unknown
        :raise ModeTypeImproperlyConfigured: when the mode's type is known but
                                             there is no information for
                                             parameters (if and when they are
                                             required by the mode)
        :return: a tuple with two values: the mode type and if it requires
                 a parameter

        ::

            >>> chanmodes = {'A': tuple('beI'), 'B': tuple('k')}
            >>> t_params = {
            ...     'A': ParamRequired.ALWAYS,
            ...     'B': ParamRequired.ADDED,
            ... }
            >>> mm = ModeParser(chanmodes, t_params)
            >>> mm.get_mode_info('e', False)
            ('A', True)
            >>> mm.get_mode_info('k', False)
            ('B', False)
            >>> mm.get_mode_info('e', True)
            ('A', True)
            >>> mm.get_mode_info('k', True)
            ('B', True)

        .. note::

            A user privilege ``mode`` doesn't have a type and will trigger
            a :exc:`ModeTypeUnknown` exception.
        """
        letter = self.get_mode_type(mode)

        if letter not in self.type_params:
            # we don't know how to handle this type of mode
            raise ModeTypeImproperlyConfigured(mode, letter)

        type_param = self.type_params[letter]
        return letter, not type_param == ParamRequired.NEVER and (
            type_param == ParamRequired.ALWAYS
            or type_param == ParamRequired.ADDED and is_added
            or type_param == ParamRequired.REMOVED and not is_added
        )

    def parse(self, modestring: str, params: tuple[str, ...]) -> ModeMessage:
        """Parse a ``modestring`` for a channel with its ``params``.

        :param modestring: suite of modes with +/- sign, such as ``+b-v``
        :param params: tuple of parameters as given by the MODE message
        :return: the parsed and validated information for that ``modestring``

        This method parses a modestring, i.e. a suite of modes and privileges
        with + and - signs. The result is a :class:`ModeMessage` with:

        * parsed modes, with their parameters when required
        * parsed privileges, with their parameters
        * ignore modes (unknown and invalid modes)
        * leftover parameters (parameter unused)

        For example this message:

        .. code-block:: irc

            :irc.example.com MODE #foobar -o+vi mario luigi bowser

        Should be parsed like this::

            >>> modestring = '-o+vi'
            >>> params = ('mario', 'luigi', 'bowser')
            >>> modes = modeparser.parse(modestring, params)
            >>> modes.modes
            (('D', 'i', True, None),)
            >>> modes.privileges
            (('o', False, 'mario'), ('v', True, 'luigi'))
            >>> modes.leftover_params
            ('bowser',)

        The modestring ``-o+vi`` means::

        * remove ``o`` privileges to user ``mario``
        * add ``v`` privileges to user ``luigi``
        * set ``i`` mode on channel ``#foobar`` (no parameter required)

        Which means that ``bowser`` shouldn't be here, and can be retrieved
        through the ``leftover_params`` attribute.
        """
        imodes = iter(parse_modestring(modestring))
        iparams = iter(params)
        modes: list[ModeDetails] = []
        privileges: list[PrivilegeDetails] = []

        for mode, is_added in imodes:
            required = False

            try:
                if mode in self.privileges:
                    priv_param: str = next(iparams)
                    privileges.append((mode, is_added, priv_param))
                else:
                    mode_param: Optional[str] = None
                    letter, required = self.get_mode_info(mode, is_added)
                    if required:
                        mode_param = next(iparams)
                    modes.append((letter, mode, is_added, mode_param))
            except StopIteration:
                # Not enough parameters: we have to stop here
                return ModeMessage(
                    tuple(modes),
                    tuple(privileges),
                    ((mode, is_added),) + tuple(imodes),
                    tuple(),
                )
            except ModeException as modeerror:
                LOGGER.debug(
                    'Invalid modestring: %r; error: %s',
                    modestring,
                    modeerror,
                )
                return ModeMessage(
                    tuple(modes),
                    tuple(privileges),
                    ((mode, is_added),) + tuple(imodes),
                    tuple(iparams),
                )

        return ModeMessage(
            tuple(modes),
            tuple(privileges),
            tuple(imodes),
            tuple(iparams),
        )



""":mod:`sopel.irc.utils` contains low-level tools for IRC protocol handling.

.. warning::

    This is all internal code, not intended for direct use by plugins. It is
    subject to change between versions, even patch releases, without any
    advance notice.

"""
# Copyright 2019, Florian Strzelecki <florian.strzelecki@gmail.com>
#
# Licensed under the Eiffel Forum License 2.
from __future__ import annotations

from typing import NamedTuple

from sopel.lifecycle import deprecated


def safe(string: str) -> str:
    """Remove disallowed bytes from a string, and ensure Unicode.

    :param string: input text to process
    :return: the string as Unicode without characters prohibited in IRC messages
    :raises TypeError: when ``string`` is ``None``

    This function removes newlines and null bytes from a string. It will always
    return a Unicode ``str``, even if given non-Unicode input, but doesn't strip
    or alter the string in any other way::

        >>> safe('some \\x00text\\r\\n')
        'some text'

    This is useful to ensure a string can be used in a IRC message. Parameters
    can **never** contain NUL, CR, or LF octets, per :rfc:`2812#section-2.3.1`.

    .. versionchanged:: 7.1

        This function now raises a :exc:`TypeError` instead of an unpredictable
        behaviour when given ``None``.

    .. versionchanged:: 8.0.1

        Also remove NUL (``\\x00``) in addition to CR/LF.

    """
    if string is None:
        raise TypeError('safe function requires a string, not NoneType')
    if isinstance(string, bytes):
        string = string.decode("utf8")
    string = string.replace('\n', '')
    string = string.replace('\r', '')
    string = string.replace('\x00', '')
    return string


@deprecated('CapReq is obsolete.', version='8.0', removed_in='9.0')
class CapReq:
    """Obsolete representation of a CAP REQ.

    .. deprecated:: 8.0

        This class is deprecated. See :class:`sopel.plugin.capability` instead.

        This will be removed in Sopel 9.

    """
    def __init__(self, prefix, plugin, failure=None, arg=None, success=None):
        def nop(bot, cap):
            pass
        self.prefix = prefix
        self.plugin = plugin
        self.arg = arg
        self.failure = failure or nop
        self.success = success or nop


class MyInfo(NamedTuple):
    """Store client, servername, and version from ``RPL_MYINFO`` events.

    .. seealso::

        https://modern.ircdocs.horse/#rplmyinfo-004

    """
    client: str
    servername: str
    version: str



"""Core Sopel plugin that handles IRC protocol functions.

This plugin allows the bot to run without user-facing functionality:

* it handles client capability negotiation
* it handles client auth (both nick auth and server auth)
* it handles connection registration (RPL_WELCOME, RPL_LUSERCLIENT), dealing
  with error cases such as nick already in use
* it tracks known channels & users (join, quit, nick change and other events)
* it manages blocked (ignored) users

This is written as a plugin to make it easier to extend to support more
responses to standard IRC codes without having to shove them all into the
dispatch function in :class:`sopel.bot.Sopel` and making it easier to maintain.
"""
# Copyright 2008-2011, Sean B. Palmer (inamidst.com) and Michael Yanovich
# (yanovich.net)
# Copyright © 2012, Elad Alfassa <elad@fedoraproject.org>
# Copyright 2012-2015, Elsie Powell embolalia.com
# Copyright 2019, Florian Strzelecki <florian.strzelecki@gmail.com>
#
# Licensed under the Eiffel Forum License 2.
from __future__ import annotations

import base64
import collections
import copy
from datetime import datetime, timedelta, timezone
import functools
import logging
import re
import time
from typing import Callable, Optional, TYPE_CHECKING

from sopel import config, plugin
from sopel.irc import isupport, utils
from sopel.tools import events, jobs, SopelMemory, target

if TYPE_CHECKING:
    from sopel.bot import Sopel, SopelWrapper
    from sopel.tools import Identifier
    from sopel.trigger import Trigger


LOGGER = logging.getLogger(__name__)

WHOX_QUERY = '%nuachrtf'
"""List of WHOX flags coretasks requests."""
WHOX_QUERYTYPE = '999'
"""WHOX querytype to indicate requests/responses from coretasks.

Other plugins should use a different querytype.
"""

MODE_PREFIX_PRIVILEGES = {
    "v": plugin.VOICE,
    "h": plugin.HALFOP,
    "o": plugin.OP,
    "a": plugin.ADMIN,
    "q": plugin.OWNER,
    "y": plugin.OPER,
    "Y": plugin.OPER,
}


def _handle_account_and_extjoin_capabilities(
    cap_req: tuple[str, ...], bot: SopelWrapper, acknowledged: bool,
) -> plugin.CapabilityNegotiation:
    if acknowledged:
        return plugin.CapabilityNegotiation.DONE

    name = ', '.join(cap_req)
    owner_account = bot.settings.core.owner_account
    admin_accounts = bot.settings.core.admin_accounts

    LOGGER.info(
        'Server does not support "%s". '
        'User account validation unavailable or limited.',
        name,
    )
    if owner_account or admin_accounts:
        LOGGER.warning(
            'Owner or admin accounts are configured, but "%s" is not '
            'supported by the server. This may cause unexpected behavior.',
            name,
        )

    return plugin.CapabilityNegotiation.DONE


def _handle_sasl_capability(
    cap_req: tuple[str, ...], bot: SopelWrapper, acknowledged: bool,
) -> plugin.CapabilityNegotiation:
    # Manage CAP REQ :sasl
    auth_method = bot.settings.core.auth_method
    server_auth_method = bot.settings.core.server_auth_method
    is_required = 'sasl' in (auth_method, server_auth_method)

    if not is_required:
        # not required: we are fine, available or not
        return plugin.CapabilityNegotiation.DONE
    elif not acknowledged:
        # required but not available: error, we must stop here
        LOGGER.error(
            'SASL capability is not enabled; '
            'cannot authenticate with SASL.',
        )
        return plugin.CapabilityNegotiation.ERROR

    # Check SASL configuration (password is required for PLAIN/SCRAM)
    password, mech = _get_sasl_pass_and_mech(bot)
    if mech != "EXTERNAL" and not password:
        raise config.ConfigurationError(
            'SASL authentication required but no password available; '
            'please check your configuration file.',
        )

    cap_info = bot.capabilities.get_capability_info('sasl')
    cap_params = cap_info.params

    available_mechs = cap_params.split(',') if cap_params else []

    if available_mechs and mech not in available_mechs:
        # Raise an error if configured to use an unsupported SASL mechanism,
        # but only if the server actually advertised supported mechanisms,
        # i.e. this network supports SASL 3.2

        # SASL 3.1 failure is handled (when possible)
        # by the sasl_mechs() function

        # See https://github.com/sopel-irc/sopel/issues/1780 for background
        raise config.ConfigurationError(
            'SASL mechanism "{mech}" is not advertised by this server; '
            'available mechanisms are: {available}.'.format(
                mech=mech,
                available=', '.join(available_mechs),
            )
        )

    bot.write(('AUTHENTICATE', mech))

    # If we want to do SASL, we have to wait before we can send CAP END. So if
    # we are, wait on 903 (SASL successful) to send it.
    return plugin.CapabilityNegotiation.CONTINUE


CAP_ECHO_MESSAGE = plugin.capability('echo-message')
CAP_MULTI_PREFIX = plugin.capability('multi-prefix')
CAP_AWAY_NOTIFY = plugin.capability('away-notify')
CAP_CHGHOST = plugin.capability('chghost')
CAP_CAP_NOTIFY = plugin.capability('cap-notify')
CAP_SERVER_TIME = plugin.capability('server-time')
CAP_USERHOST_IN_NAMES = plugin.capability('userhost-in-names')
CAP_MESSAGE_TAGS = plugin.capability('message-tags')
CAP_ACCOUNT_NOTIFY = plugin.capability(
    'account-notify', handler=_handle_account_and_extjoin_capabilities)
CAP_EXTENDED_JOIN = plugin.capability(
    'extended-join', handler=_handle_account_and_extjoin_capabilities)
CAP_ACCOUNT_TAG = plugin.capability(
    'account-tag', handler=_handle_account_and_extjoin_capabilities)
CAP_SASL = plugin.capability('sasl', handler=_handle_sasl_capability)


def setup(bot: Sopel) -> None:
    """Set up the coretasks plugin.

    The setup phase is used to activate the throttle feature to prevent a flood
    of JOIN commands when there are too many channels to join.
    """
    bot.memory['retry_join'] = SopelMemory()
    bot.memory['join_events_queue'] = collections.deque()

    # Manage JOIN flood protection
    if bot.settings.core.throttle_join:
        wait_interval = max(bot.settings.core.throttle_wait, 1)
        job = jobs.Job(
            [wait_interval],
            plugin='coretasks',
            label='throttle_join',
            handler=_join_event_processing,
            threaded=True,
            doc=None,
        )
        bot.scheduler.register(job)


def shutdown(bot):
    """Clean up coretasks-related values in the bot's memory."""
    bot.memory['retry_join'] = SopelMemory()
    try:
        bot.memory['join_events_queue'].clear()
    except KeyError:
        pass


def _join_event_processing(bot):
    """Process a batch of JOIN event from the ``join_events_queue`` queue.

    Every time this function is executed, it processes at most
    ``throttle_join`` JOIN events. For each JOIN, it sends a WHO request to
    know more about the channel. This will prevent an excess of flood when
    there are too many channels to join at once.
    """
    batch_size = max(bot.settings.core.throttle_join, 1)
    for _ in range(batch_size):
        try:
            channel = bot.memory['join_events_queue'].popleft()
        except IndexError:
            break
        LOGGER.debug("Sending MODE and WHO after channel JOIN: %s", channel)
        bot.write(["MODE", channel])
        _send_who(bot, channel)


def auth_after_register(bot):
    """Do NickServ/AuthServ auth.

    :param bot: a connected Sopel instance
    :type bot: :class:`sopel.bot.Sopel`

    This function can be used, **after** the bot is connected, to handle one of
    these auth methods:

    * ``nickserv``: send a private message to the NickServ service
    * ``authserv``: send an ``AUTHSERV`` command
    * ``Q``: send an ``AUTH`` command
    * ``userserv``: send a private message to the UserServ service

    .. important::

        If ``core.auth_method`` is set, then ``core.nick_auth_method`` will be
        ignored. If none is set, then this function does nothing.

    """
    if bot.config.core.auth_method:
        auth_method = bot.config.core.auth_method
        auth_username = bot.config.core.auth_username
        auth_password = bot.config.core.auth_password
        auth_target = bot.config.core.auth_target
    elif bot.config.core.nick_auth_method:
        auth_method = bot.config.core.nick_auth_method
        auth_username = (bot.config.core.nick_auth_username or
                         bot.config.core.nick)
        auth_password = bot.config.core.nick_auth_password
        auth_target = bot.config.core.nick_auth_target
    else:
        return

    # nickserv-based auth method needs to check for current nick
    if auth_method == 'nickserv':
        if bot.nick != bot.make_identifier(bot.settings.core.nick):
            LOGGER.warning("Sending nickserv GHOST command.")
            bot.say(
                'GHOST %s %s' % (bot.settings.core.nick, auth_password),
                auth_target or 'NickServ')
        else:
            bot.say('IDENTIFY %s' % auth_password, auth_target or 'NickServ')

    # other methods use account instead of nick
    elif auth_method == 'authserv':
        bot.write(('AUTHSERV', 'auth', auth_username, auth_password))
    elif auth_method == 'Q':
        bot.write(('AUTH', auth_username, auth_password))
    elif auth_method == 'userserv':
        bot.say("LOGIN %s %s" % (auth_username, auth_password),
                auth_target or 'UserServ')


def _execute_perform(bot):
    """Execute commands specified to perform on IRC server connect.

    This function executes the list of commands that can be found in the
    ``core.commands_on_connect`` setting. It automatically replaces any
    ``$nickname`` placeholder in the command with the bot's configured nick.
    """
    if not bot.connection_registered:
        # How did you even get this command, bot?
        raise Exception('Bot must be connected to server to perform commands.')

    commands = bot.config.core.commands_on_connect
    count = len(commands)

    if not count:
        LOGGER.info("No custom command to execute.")
        return

    LOGGER.info("Executing %d custom commands.", count)
    for i, command in enumerate(commands, 1):
        command = command.replace('$nickname', bot.config.core.nick)
        LOGGER.debug("Executing custom command [%d/%d]: %s", i, count, command)
        bot.write((command,))


@plugin.event(events.ERR_NICKNAMEINUSE)
@plugin.thread(False)
@plugin.unblockable
@plugin.priority('medium')
def on_nickname_in_use(bot, trigger):
    """Change the bot's nick when the current one is already in use.

    This can be triggered when the bot disconnects then reconnects before the
    server can notice a client timeout. Other reasons include mischief,
    trolling, and obviously, PEBKAC.

    This will change the current nick by adding a trailing ``_``. If the bot
    sees that a user with its configured nick disconnects (see ``QUIT`` event
    handling), the bot will try to regain it.
    """
    LOGGER.error(
        "Nickname already in use! (Nick: %s; Sender: %s; Args: %r)",
        trigger.nick,
        trigger.sender,
        trigger.args,
    )
    bot.change_current_nick(bot.nick + '_')


@plugin.require_privmsg("This command only works as a private message.")
@plugin.require_admin("This command requires admin privileges.")
@plugin.commands('execute')
def execute_perform(bot, trigger):
    """Execute commands specified to perform on IRC server connect.

    This allows a bot owner or admin to force the execution of commands
    that are automatically performed when the bot connects.
    """
    _execute_perform(bot)


@plugin.event(events.RPL_WELCOME, events.RPL_LUSERCLIENT)
@plugin.thread(False)
@plugin.unblockable
@plugin.priority('medium')
def startup(bot, trigger):
    """Do tasks related to connecting to the network.

    ``001 RPL_WELCOME`` is from RFC2812 and is the first message that is sent
    after the connection has been registered on the network.

    ``251 RPL_LUSERCLIENT`` is a mandatory message that is sent after the
    client connects to the server in RFC1459. RFC2812 does not require it and
    all networks might not send it. We support both.

    If ``sopel.irc.AbstractBot.connection_registered`` is set, this function
    does nothing and returns immediately. Otherwise, the flag is set and the
    function proceeds normally to:

    1. trigger auth method
    2. set bot's ``MODE`` (from ``core.modes``)
    3. join channels (or queue them to join later)
    4. check for security when the ``account-tag`` capability is enabled
    5. execute custom commands
    """
    if bot.connection_registered:
        return

    LOGGER.info(
        'Enabled client capabilities: %s',
        ', '.join(bot.capabilities.enabled),
    )

    # nick shenanigans are serious business, but fortunately RPL_WELCOME
    # includes the actual nick used by the server after truncation, removal
    # of invalid characters, etc. so we can check for such shenanigans
    if trigger.event == events.RPL_WELCOME:
        if bot.nick != trigger.args[0]:
            # setting modes below is just one of the things that won't work
            # as expected if the conditions for running this block are met
            privmsg = (
                "Hi, I'm your bot, %s. The IRC server didn't assign me the "
                "nick you configured. This can cause problems for me, and "
                "make me do weird things. You'll probably want to stop me, "
                "figure out why my nick isn't acceptable, and fix that before "
                "starting me again." % bot.nick
            )
            debug_msg = (
                "RPL_WELCOME indicated the server did not accept the bot's "
                "configured nickname. Requested '%s'; got '%s'. This can "
                "cause unexpected behavior. Please modify the configuration "
                "and restart the bot." % (bot.nick, trigger.args[0])
            )
            LOGGER.critical(debug_msg)
            bot.say(privmsg, bot.config.core.owner)

    # set flag
    bot._connection_registered.set()

    # handle auth method
    auth_after_register(bot)

    # set bot's MODE
    modes = bot.config.core.modes
    if modes:
        if not modes.startswith(('+', '-')):
            # Assume "+" by default.
            modes = '+' + modes
        bot.write(('MODE', bot.nick, modes))

    # warn for insecure auth method if necessary
    if (
        not bot.config.core.owner_account and
        bot.capabilities.is_enabled('account-tag') and
        '@' not in bot.config.core.owner
    ):
        msg = (
            "This network supports using network services to identify you as "
            "my owner, rather than just matching your nickname. This is much "
            "more secure. If you'd like to do this, make sure you're logged in "
            "and reply with \"{}useserviceauth\""
        ).format(bot.config.core.help_prefix)
        bot.say(msg, bot.config.core.owner)

    # execute custom commands
    _execute_perform(bot)


@plugin.event(events.RPL_ISUPPORT)
@plugin.thread(False)
@plugin.unblockable
@plugin.rule('are supported by this server')
@plugin.priority('medium')
def handle_isupport(bot, trigger):
    """Handle ``RPL_ISUPPORT`` events."""
    # remember if certain actionable tokens are known to be supported,
    # before parsing RPL_ISUPPORT
    botmode_support = 'BOT' in bot.isupport
    namesx_support = 'NAMESX' in bot.isupport
    uhnames_support = 'UHNAMES' in bot.isupport
    casemapping_support = 'CASEMAPPING' in bot.isupport
    chantypes_support = 'CHANTYPES' in bot.isupport

    # parse ISUPPORT message from server
    parameters = {}
    for arg in trigger.args:
        try:
            key, value = isupport.parse_parameter(arg)
            parameters[key] = value
        except ValueError:
            # ignore malformed parameter: log a warning and continue
            LOGGER.warning("Unable to parse ISUPPORT parameter: %r", arg)

    bot._isupport = bot._isupport.apply(**parameters)

    # update bot's mode parser
    if 'CHANMODES' in bot.isupport:
        bot.modeparser.chanmodes = bot.isupport.CHANMODES

    if 'PREFIX' in bot.isupport:
        bot.modeparser.privileges = set(bot.isupport.PREFIX.keys())

    # rebuild nick when CASEMAPPING and/or CHANTYPES are set
    if any((
        # was CASEMAPPING support status updated?
        not casemapping_support and 'CASEMAPPING' in bot.isupport,
        # was CHANTYPES support status updated?
        not chantypes_support and 'CHANTYPES' in bot.isupport,
    )):
        # these parameters change how the bot makes Identifiers
        # since bot.nick is an Identifier, it must be rebuilt
        bot.rebuild_nick()

    # was BOT mode support status updated?
    if not botmode_support and 'BOT' in bot.isupport:
        # yes it was! set our mode unless the config overrides it
        botmode = bot.isupport['BOT']
        modes_setting = bot.config.core.modes

        if not modes_setting or botmode not in bot.config.core.modes:
            bot.write(('MODE', bot.nick, '+' + botmode))

    # was NAMESX support status updated?
    if not namesx_support and 'NAMESX' in bot.isupport:
        # yes it was!
        if not bot.capabilities.is_enabled('multi-prefix'):
            # and the multi-prefix capability is not enabled
            # so we can ask the server to use the NAMESX feature
            bot.write(('PROTOCTL', 'NAMESX'))

    # was UHNAMES support status updated?
    if not uhnames_support and 'UHNAMES' in bot.isupport:
        # yes it was!
        if not bot.capabilities.is_enabled('userhost-in-names'):
            # and the userhost-in-names capability is not enabled
            # so we should ask for UHNAMES instead
            bot.write(('PROTOCTL', 'UHNAMES'))


@plugin.event(events.RPL_ENDOFMOTD, events.ERR_NOMOTD)
@plugin.thread(False)
@plugin.unblockable
@plugin.priority('medium')
def join_channels(bot, trigger):
    # join channels
    channels = bot.config.core.channels
    if not channels:
        LOGGER.info("No initial channels to JOIN.")

    elif bot.config.core.throttle_join:
        throttle_rate = int(bot.config.core.throttle_join)
        throttle_wait = max(bot.config.core.throttle_wait, 1)
        channels_joined = 0

        LOGGER.info(
            "Joining %d channels (with JOIN throttle ON); "
            "this may take a moment.",
            len(channels))

        for channel in channels:
            channels_joined += 1
            if not channels_joined % throttle_rate:
                LOGGER.debug(
                    "Waiting %ds before next JOIN batch.",
                    throttle_wait)
                time.sleep(throttle_wait)
            bot.join(channel)
    else:
        LOGGER.info(
            "Joining %d channels (with JOIN throttle OFF); "
            "this may take a moment.",
            len(channels))

        for channel in bot.config.core.channels:
            bot.join(channel)


@plugin.event(events.RPL_MYINFO)
@plugin.thread(False)
@plugin.unblockable
@plugin.priority('medium')
def parse_reply_myinfo(bot, trigger):
    """Handle ``RPL_MYINFO`` events."""
    # keep <client> <servername> <version> only
    # the trailing parameters (mode types) should be read from ISUPPORT
    bot._myinfo = utils.MyInfo(*trigger.args[0:3])

    LOGGER.info(
        "Received RPL_MYINFO from server: %s, %s, %s",
        bot._myinfo.client,
        bot._myinfo.servername,
        bot._myinfo.version,
    )


@plugin.require_privmsg()
@plugin.require_owner()
@plugin.commands('useserviceauth')
def enable_service_auth(bot, trigger):
    """Set owner's account from an authenticated owner.

    This command can be used to automatically configure ``core.owner_account``
    when the owner is known and has a registered account, but the bot doesn't
    have ``core.owner_account`` configured.

    This doesn't work if the ``account-tag`` capability is not available.
    """
    if bot.config.core.owner_account:
        return
    if not bot.capabilities.is_enabled('account-tag'):
        bot.say('This server does not fully support services auth, so this '
                'command is not available.')
        return
    if not trigger.account:
        bot.say('You must be logged in to network services before using this '
                'command.')
        return
    bot.config.core.owner_account = trigger.account
    bot.config.save()
    bot.say('Success! I will now use network services to identify you as my '
            'owner.')
    LOGGER.info(
        "User %s set %s as owner account.",
        trigger.nick,
        trigger.account,
    )


@plugin.event(events.ERR_NOCHANMODES)
@plugin.priority('medium')
def retry_join(bot, trigger):
    """Give NickServ enough time to identify on a +R channel.

    Give NickServ enough time to identify, and retry rejoining an
    identified-only (+R) channel. Maximum of ten rejoin attempts.
    """
    channel = trigger.args[1]
    if channel in bot.memory['retry_join'].keys():
        bot.memory['retry_join'][channel] += 1
        if bot.memory['retry_join'][channel] > 10:
            LOGGER.warning("Failed to join %s after 10 attempts.", channel)
            return
        LOGGER.info(
            "Rejoining channel %r failed, will retry in 6s.",
            str(channel))
        time.sleep(6)
    else:
        bot.memory['retry_join'][channel] = 0

    attempt = bot.memory['retry_join'][channel] + 1
    LOGGER.info(
        "Trying to rejoin channel %r (attempt %d/10)",
        str(channel), attempt)
    bot.join(channel)


@plugin.rule('(.*)')
@plugin.event(events.RPL_NAMREPLY)
@plugin.thread(False)
@plugin.unblockable
@plugin.priority('medium')
def handle_names(bot, trigger):
    """Handle NAMES responses.

    This function keeps track of users' privileges when Sopel joins channels.
    """
    # TODO specific to one channel type. See issue 281.
    channels = re.search(r'(#\S*)', trigger.raw)
    if not channels:
        return
    channel = bot.make_identifier(channels.group(1))
    if channel not in bot.channels:
        bot.channels[channel] = target.Channel(
            channel,
            identifier_factory=bot.make_identifier,
        )

    # This could probably be made flexible in the future, but I don't think
    # it'd be worth it.
    # If this ever needs to be updated, remember to change the mode handling in
    # the WHO-handler functions below, too.
    mapping = {
        "+": plugin.VOICE,
        "%": plugin.HALFOP,
        "@": plugin.OP,
        "&": plugin.ADMIN,
        "~": plugin.OWNER,
        "!": plugin.OPER,
    }

    uhnames = 'UHNAMES' in bot.isupport
    userhost_in_names = bot.capabilities.is_enabled('userhost-in-names')

    names = trigger.split()
    for name in names:
        username = hostname = None

        if uhnames or userhost_in_names:
            try:
                name, mask = name.rsplit('!', 1)
                username, hostname = mask.split('@', 1)
            except ValueError:
                # server advertised either UHNAMES or userhost-in-names, but
                # isn't sending the hostmask with all listed nicks
                # It's probably ZNC. https://github.com/znc/znc/issues/1224
                LOGGER.debug(
                    '%s is enabled, but still got RPL_NAMREPLY item without a hostmask. '
                    'IRC server/bouncer is not spec compliant.',
                    'UHNAMES' if uhnames else 'userhost-in-names')

        priv = 0
        for prefix, value in mapping.items():
            if prefix in name:
                priv = priv | value

        nick = bot.make_identifier(name.lstrip(''.join(mapping.keys())))
        user = bot.users.get(nick)
        if user is None:
            # The username/hostname will be included in a NAMES reply only if
            # userhost-in-names is available. We can use them if present.
            # Fortunately, the user should already exist in bot.users by the
            # time this code runs, so this is 99.9% ass-covering.
            user = target.User(nick, username, hostname)
            bot.users[nick] = user
        bot.channels[channel].add_user(user, privs=priv)


@plugin.rule('(.*)')
@plugin.event('MODE')
@plugin.thread(False)
@plugin.unblockable
@plugin.priority('medium')
def track_modes(bot, trigger):
    """Track changes from received MODE commands."""
    _parse_modes(bot, trigger.args)


@plugin.priority('high')
@plugin.event(events.RPL_CHANNELMODEIS)
@plugin.thread(False)
@plugin.unblockable
def initial_modes(bot, trigger):
    """Populate channel modes from response to MODE request sent after JOIN."""
    _parse_modes(bot, trigger.args[1:], clear=True)


def _parse_modes(bot, args, clear=False):
    """Parse MODE message and apply changes to internal state.

    Sopel, by default, doesn't know how to parse other types than A, B, C, and
    D, and only a preset of privileges.

    .. seealso::

        Parsing mode messages can be tricky and complicated to understand. In
        any case it is better to read the IRC specifications about channel
        modes at https://modern.ircdocs.horse/#channel-mode

    """
    channel_name = bot.make_identifier(args[0])
    if channel_name.is_nick():
        # We don't do anything with user modes
        LOGGER.debug("Ignoring user modes: %r", args)
        return

    channel = bot.channels[channel_name]

    # Unreal 3 sometimes sends an extraneous trailing space. If we're short an
    # arg, we'll find out later.
    if args[-1] == "":
        args.pop()
    # If any args are still empty, that's something we may not be prepared for,
    # but let's continue anyway hoping they're trailing / not important.
    if len(args) < 2 or not all(args):
        LOGGER.debug(
            "The server sent a possibly malformed MODE message: %r", args)

    # parse the modestring with the parameters
    modeinfo = bot.modeparser.parse(args[1], tuple(args[2:]))

    # set, unset, or update channel's modes based on the mode type
    # modeinfo.modes contains only the valid parsed modes
    # coretask can handle type A, B, C, and D only
    modes = {} if clear else copy.deepcopy(channel.modes)
    for letter, mode, is_added, param in modeinfo.modes:
        if letter == 'A':
            # type A is a multi-value mode and always requires a parameter
            if mode not in modes:
                modes[mode] = set()
            if is_added:
                modes[mode].add(param)
            elif param in modes[mode]:
                modes[mode].remove(param)
                # remove mode if empty
                if not modes[mode]:
                    modes.pop(mode)
        elif letter == 'B':
            # type B is a single-value mode and always requires a parameter
            if is_added:
                modes[mode] = param
            elif mode in modes:
                modes.pop(mode)
        elif letter == 'C':
            # type C is a single-value mode and requires a parameter when added
            if is_added:
                modes[mode] = param
            elif mode in modes:
                modes.pop(mode)
        elif letter == 'D':
            # type D is a flag (True or False) and doesn't have a parameter
            if is_added:
                modes[mode] = True
            elif mode in modes:
                modes.pop(mode)

    # atomic change of channel's modes
    channel.modes = modes

    # update user privileges in channel
    # modeinfo.privileges contains only the valid parsed privileges
    for privilege, is_added, param in modeinfo.privileges:
        # User privs modes, always have a param
        nick = bot.make_identifier(param)
        priv = channel.privileges.get(nick, 0)
        value = MODE_PREFIX_PRIVILEGES[privilege]
        if is_added:
            priv = priv | value
        else:
            priv = priv & ~value
        channel.privileges[nick] = priv

    # log ignored modes (modes Sopel doesn't know how to handle)
    if modeinfo.ignored_modes:
        LOGGER.warning(
            "Unknown MODE message, sending WHO. Message was: %r",
            args,
        )
        # send a WHO message to ensure we didn't miss anything
        _send_who(bot, channel_name)

    # log leftover parameters (too many arguments)
    if modeinfo.leftover_params:
        LOGGER.warning(
            "Too many arguments received for MODE: args=%r chanmodes=%r",
            args,
            bot.modeparser.chanmodes,
        )

    LOGGER.info("Updated mode for channel: %s", channel.name)
    LOGGER.debug("Channel %r mode: %r", str(channel.name), channel.modes)


@plugin.event('NICK')
@plugin.thread(False)
@plugin.unblockable
@plugin.priority('medium')
def track_nicks(bot, trigger):
    """Track nickname changes and maintain our chanops list accordingly."""
    old = trigger.nick
    new = bot.make_identifier(trigger)

    # Give debug message, and PM the owner, if the bot's own nick changes.
    if old == bot.nick and new != bot.nick:
        # Is this the original nick being regained?
        # e.g. by ZNC's keepnick module running in front of Sopel
        if old != bot.config.core.nick and new == bot.config.core.nick:
            LOGGER.info(
                "Regained configured nick. Restarting is still recommended.")
        else:
            privmsg = (
                "Hi, I'm your bot, %s. Something has made my nick change. This "
                "can cause some problems for me, and make me do weird things. "
                "You'll probably want to restart me, and figure out what made "
                "that happen so you can stop it happening again. (Usually, it "
                "means you tried to give me a nick that's protected by NickServ.)"
            ) % bot.config.core.nick
            debug_msg = (
                "Nick changed by server. This can cause unexpected behavior. "
                "Please restart the bot."
            )
            LOGGER.critical(debug_msg)
            bot.say(privmsg, bot.config.core.owner)

        # Always update bot.nick anyway so Sopel doesn't lose its self-identity.
        # This should cut down the number of "weird things" that happen while
        # the active nick doesn't match the config, but it's not a substitute
        # for regaining the expected nickname.
        LOGGER.info("Updating bot.nick property with server-changed nick.")
        bot._nick = new
        return

    for channel in bot.channels.values():
        channel.rename_user(old, new)
    if old in bot.users:
        bot.users[new] = bot.users.pop(old)

    LOGGER.info("User named %r is now known as %r.", str(old), str(new))


@plugin.rule('(.*)')
@plugin.event('PART')
@plugin.thread(False)
@plugin.unblockable
@plugin.priority('medium')
def track_part(bot, trigger):
    """Track users leaving channels."""
    nick = trigger.nick
    channel = trigger.sender
    _remove_from_channel(bot, nick, channel)
    LOGGER.info("User %r left a channel: %s", str(nick), channel)


@plugin.event('KICK')
@plugin.thread(False)
@plugin.unblockable
@plugin.priority('medium')
def track_kick(bot, trigger):
    """Track users kicked from channels."""
    nick = bot.make_identifier(trigger.args[1])
    channel = trigger.sender
    _remove_from_channel(bot, nick, channel)
    LOGGER.info(
        "User %r got kicked by %r from a channel: %s",
        str(nick),
        str(trigger.nick),
        channel,
    )


def _remove_from_channel(bot, nick, channel):
    if nick == bot.nick:
        bot.channels.pop(channel, None)

        lost_users = []
        for nick_, user in bot.users.items():
            user.channels.pop(channel, None)
            if not user.channels:
                lost_users.append(nick_)
        for nick_ in lost_users:
            bot.users.pop(nick_, None)
    else:
        user = bot.users.get(nick)
        if user and channel in user.channels:
            bot.channels[channel].clear_user(nick)
            if not user.channels:
                bot.users.pop(nick, None)


def _send_who(bot, mask):
    if 'WHOX' in bot.isupport:
        # WHOX syntax, see http://faerion.sourceforge.net/doc/irc/whox.var
        # Needed for accounts in WHO replies. The `WHOX_QUERYTYPE` parameter
        # for WHO is used to identify the reply from the server and confirm
        # that it has the requested format. WHO replies with different
        # querytypes in the response were initiated elsewhere and will be
        # ignored.
        bot.write(['WHO', mask, '{},{}'.format(WHOX_QUERY, WHOX_QUERYTYPE)])
    else:
        # We might be on an old network, but we still care about keeping our
        # user list updated
        bot.write(['WHO', mask])

    target_id = bot.make_identifier(mask)
    if not target_id.is_nick():
        bot.channels[target_id].last_who = datetime.now(timezone.utc)


@plugin.interval(30)
def _periodic_send_who(bot):
    """Periodically send a WHO request to keep user information up-to-date."""
    if bot.capabilities.is_enabled('away-notify'):
        # WHO not needed to update 'away' status
        return

    # Loop through the channels to find the one that has the longest time since the last WHO
    # request, and issue a WHO request only if the last request for the channel was more than
    # 120 seconds ago.
    who_trigger_time = datetime.now(timezone.utc) - timedelta(seconds=120)
    selected_channel = None
    for channel_name, channel in bot.channels.items():
        if channel.last_who is None:
            # WHO was never sent yet to this channel: stop here
            selected_channel = channel_name
            break
        if channel.last_who < who_trigger_time:
            # this channel's last who request is the most outdated one at the moment
            selected_channel = channel_name
            who_trigger_time = channel.last_who

    if selected_channel is not None:
        # selected_channel's last who is either none or the oldest valid
        LOGGER.debug("Sending WHO for channel: %s", selected_channel)
        _send_who(bot, selected_channel)


@plugin.event('JOIN')
@plugin.thread(False)
@plugin.unblockable
@plugin.priority('medium')
def track_join(bot, trigger):
    """Track users joining channels.

    When a user joins a channel, the bot will send (or queue) a ``WHO`` command
    to know more about said user (privileges, modes, etc.).
    """
    channel = trigger.sender
    new_channel = channel not in bot.channels
    self_join = trigger.nick == bot.nick
    new_user = trigger.nick not in bot.users

    # is it a new channel?
    if new_channel:
        bot.channels[channel] = target.Channel(
            channel,
            identifier_factory=bot.make_identifier,
        )

    # did *we* just join?
    if self_join:
        LOGGER.info("Channel joined: %s", channel)
        bot.channels[channel].join_time = trigger.time
        if bot.settings.core.throttle_join:
            LOGGER.debug("JOIN event added to queue for channel: %s", channel)
            bot.memory['join_events_queue'].append(channel)
        else:
            LOGGER.debug("Send MODE and direct WHO for channel: %s", channel)
            bot.write(["MODE", channel])
            _send_who(bot, channel)
    else:
        LOGGER.info(
            "Channel %r joined by user: %s",
            str(channel), trigger.nick)

    # set initial values
    if new_user:
        user = target.User(trigger.nick, trigger.user, trigger.host)
        bot.users[trigger.nick] = user
    else:
        user = bot.users.get(trigger.nick)
    bot.channels[channel].add_user(user)

    if len(trigger.args) > 1 and trigger.args[1] != '*' and (
        bot.capabilities.is_enabled('account-notify') and
        bot.capabilities.is_enabled('extended-join')
    ):
        user.account = trigger.args[1]

    if new_user and not new_channel:
        # send WHO to populate new user's realname etc.
        _send_who(bot, trigger.nick)


@plugin.event('QUIT')
@plugin.thread(False)
@plugin.unblockable
@plugin.priority('medium')
def track_quit(bot, trigger):
    """Track when users quit channels."""
    for channel in bot.channels.values():
        channel.clear_user(trigger.nick)
    bot.users.pop(trigger.nick, None)

    LOGGER.info("User quit: %s", trigger.nick)

    configured_nick = bot.make_identifier(bot.settings.core.nick)
    if trigger.nick == configured_nick and trigger.nick != bot.nick:
        # old nick is now available, let's change nick again
        bot.change_current_nick(bot.settings.core.nick)
        auth_after_register(bot)


def _receive_cap_ls_reply(bot: SopelWrapper, trigger: Trigger) -> None:
    if not bot.capabilities.handle_ls(bot, trigger):
        # multi-line, we must wait for more
        return

    if not bot.request_capabilities():
        # Negotiation end because there is nothing to request
        LOGGER.info('No capability negotiation.')
        bot.write(('CAP', 'END'))


def _handle_cap_acknowledgement(
    bot: SopelWrapper,
    cap_req: tuple[str, ...],
    results: list[tuple[bool, Optional[plugin.CapabilityNegotiation]]],
    was_completed: bool,
) -> None:
    if any(
        callback_result[1] == plugin.CapabilityNegotiation.ERROR
        for callback_result in results
    ):
        # error: a plugin needs something and the bot cannot function properly
        LOGGER.error(
            'Capability negotiation failed for request: "%s"',
            ' '.join(cap_req),
        )
        bot.write(('CAP', 'END'))  # close negotiation now
        bot.quit('Error negotiating capabilities.')

    if not was_completed and bot.cap_requests.is_complete:
        # success: negotiation is complete and wasn't already
        LOGGER.info('Capability negotiation ended successfuly.')
        bot.write(('CAP', 'END'))  # close negotiation now


def _receive_cap_ack(bot: SopelWrapper, trigger: Trigger) -> None:
    was_completed = bot.cap_requests.is_complete
    cap_ack: tuple[str, ...] = bot.capabilities.handle_ack(bot, trigger)

    try:
        result: Optional[
            list[tuple[bool, Optional[plugin.CapabilityNegotiation]]]
        ] = bot.cap_requests.acknowledge(bot, cap_ack)
    except config.ConfigurationError as error:
        LOGGER.error(
            'Configuration error on ACK capability "%s": %s',
            ', '.join(cap_ack),
            error,
        )
        bot.write(('CAP', 'END'))  # close negotiation now
        bot.quit('Configuration error.')
        return None
    except Exception as error:
        LOGGER.exception(
            'Error on ACK capability "%s": %s',
            ', '.join(cap_ack),
            error,
        )
        bot.write(('CAP', 'END'))  # close negotiation now
        bot.quit('Error negotiating capabilities.')
        return None

    if result is None:
        # a plugin may have requested the capability without using the proper
        # interface: ignore
        return None

    _handle_cap_acknowledgement(bot, cap_ack, result, was_completed)


def _receive_cap_nak(bot: SopelWrapper, trigger: Trigger) -> None:
    was_completed = bot.cap_requests.is_complete
    cap_ack = bot.capabilities.handle_nak(bot, trigger)

    try:
        result: Optional[
            list[tuple[bool, Optional[plugin.CapabilityNegotiation]]]
        ] = bot.cap_requests.deny(bot, cap_ack)
    except config.ConfigurationError as error:
        LOGGER.error(
            'Configuration error on NAK capability "%s": %s',
            ', '.join(cap_ack),
            error,
        )
        bot.write(('CAP', 'END'))  # close negotiation now
        bot.quit('Configuration error.')
        return None
    except Exception as error:
        LOGGER.exception(
            'Error on NAK capability "%s": %s',
            ', '.join(cap_ack),
            error,
        )
        bot.write(('CAP', 'END'))  # close negotiation now
        bot.quit('Error negotiating capabilities.')
        return None

    if result is None:
        # a plugin may have requested the capability without using the proper
        # interface: ignore
        return None

    _handle_cap_acknowledgement(bot, cap_ack, result, was_completed)


def _receive_cap_new(bot: SopelWrapper, trigger: Trigger) -> None:
    cap_new = bot.capabilities.handle_new(bot, trigger)
    LOGGER.info('Capability is now available: %s', ', '.join(cap_new))
    # TODO: try to request what wasn't requested before


def _receive_cap_del(bot: SopelWrapper, trigger: Trigger) -> None:
    cap_del = bot.capabilities.handle_del(bot, trigger)
    LOGGER.info('Capability is now unavailable: %s', ', '.join(cap_del))
    # TODO: what to do when a CAP is removed? NAK callbacks?


CAP_HANDLERS: dict[str, Callable[[SopelWrapper, Trigger], None]] = {
    'LS': _receive_cap_ls_reply,  # Server is listing capabilities
    'ACK': _receive_cap_ack,  # Server is acknowledging a capability
    'NAK': _receive_cap_nak,  # Server is denying a capability
    'NEW': _receive_cap_new,  # Server is adding new capability
    'DEL': _receive_cap_del,  # Server is removing a capability
}


@plugin.event('CAP')
@plugin.thread(False)
@plugin.unblockable
@plugin.priority('medium')
def receive_cap_list(bot: SopelWrapper, trigger: Trigger) -> None:
    """Handle client capability negotiation."""
    subcommand = trigger.args[1]
    if subcommand in CAP_HANDLERS:
        handler = CAP_HANDLERS[subcommand]
        handler(bot, trigger)
    else:
        LOGGER.info('Unknown CAP subcommand received: %s', subcommand)


def send_authenticate(bot, token):
    """Send ``AUTHENTICATE`` command to server with the given ``token``.

    :param bot: instance of IRC bot that must authenticate
    :param str token: authentication token

    In case the ``token`` is more than 400 bytes, we need to split it and send
    as many ``AUTHENTICATE`` commands as needed. If the last chunk is 400 bytes
    long, we must also send a last empty command (`AUTHENTICATE +` is for empty
    line), so the server knows we are done with ``AUTHENTICATE``.

    .. seealso::

        https://ircv3.net/specs/extensions/sasl-3.1.html#the-authenticate-command

    """
    # payload is a base64 encoded token
    payload = base64.b64encode(token.encode('utf-8'))

    # split the payload into chunks of at most 400 bytes
    chunk_size = 400
    for i in range(0, len(payload), chunk_size):
        offset = i + chunk_size
        chunk = payload[i:offset]
        bot.write(('AUTHENTICATE', chunk))

    # send empty (+) AUTHENTICATE when payload's length is a multiple of 400
    if len(payload) % chunk_size == 0:
        bot.write(('AUTHENTICATE', '+'))


@plugin.event('AUTHENTICATE')
@plugin.thread(False)
@plugin.unblockable
@plugin.priority('medium')
def auth_proceed(bot, trigger):
    """Handle client-initiated SASL auth.

    If the chosen mechanism is client-first, the server sends an empty
    response (``AUTHENTICATE +``). In that case, Sopel will handle SASL auth
    that uses a token.

    .. important::

        If ``core.auth_method`` is set, then ``core.server_auth_method`` will
        be ignored. If none is set, then this function does nothing.

    """
    if bot.config.core.auth_method == 'sasl':
        mech = bot.config.core.auth_target or 'PLAIN'
    elif bot.config.core.server_auth_method == 'sasl':
        mech = bot.config.core.server_auth_sasl_mech or 'PLAIN'
    else:
        return

    if mech == 'EXTERNAL':
        if trigger.args[0] != '+':
            # not an expected response from the server; abort SASL
            token = '*'
        else:
            token = '+'

        bot.write(('AUTHENTICATE', token))
        return

    if bot.config.core.auth_method == 'sasl':
        sasl_username = bot.config.core.auth_username
        sasl_password = bot.config.core.auth_password
    elif bot.config.core.server_auth_method == 'sasl':
        sasl_username = bot.config.core.server_auth_username
        sasl_password = bot.config.core.server_auth_password
    else:
        # How did we get here? I am not good with computer
        return

    sasl_username = sasl_username or bot.nick

    if mech == 'PLAIN':
        if trigger.args[0] == '+':
            sasl_token = _make_sasl_plain_token(sasl_username, sasl_password)
            LOGGER.info("Sending SASL Auth token.")
            send_authenticate(bot, sasl_token)
            return
        else:
            # Not an expected response from the server
            LOGGER.warning(
                'Aborting SASL: unexpected server reply "%s"', trigger,
            )
            # Send `authenticate-abort` command
            # See https://ircv3.net/specs/extensions/sasl-3.1#the-authenticate-command
            bot.write(('AUTHENTICATE', '*'))
            return

    # TODO: Implement SCRAM challenges


def _make_sasl_plain_token(account, password):
    return '\x00'.join((account, account, password))


@plugin.event(events.RPL_SASLSUCCESS)
@plugin.thread(False)
@plugin.unblockable
@plugin.priority('medium')
def sasl_success(bot: SopelWrapper, trigger: Trigger) -> None:
    """Resume capability negotiation on successful SASL auth."""
    LOGGER.info("Successful SASL Auth.")
    bot.resume_capability_negotiation(CAP_SASL.cap_req, 'coretasks')


@plugin.event(events.ERR_SASLFAIL)
@plugin.event(events.ERR_SASLTOOLONG)
@plugin.event(events.ERR_SASLABORTED)
@plugin.event(events.ERR_NICKLOCKED)
@plugin.thread(False)
@plugin.unblockable
@plugin.priority('medium')
def sasl_fail(bot, trigger):
    """SASL Auth Failed: log the error and quit."""
    LOGGER.error(
        "SASL Auth Failed; check your configuration: %s",
        str(trigger))
    # negotiation done
    bot.resume_capability_negotiation(CAP_SASL.cap_req, 'coretasks')
    # quit
    bot.quit('SASL Auth Failed')


@plugin.event(events.RPL_SASLMECHS)
@plugin.thread(False)
@plugin.unblockable
@plugin.priority('low')
def sasl_mechs(bot, trigger):
    # Presumably we're only here if we said we actually *want* sasl, but still
    # check anyway in case the server glitched.
    password, mech = _get_sasl_pass_and_mech(bot)
    if not password:
        # negotiation done
        bot.resume_capability_negotiation(CAP_SASL.cap_req, 'coretasks')
        return

    supported_mechs = trigger.args[1].split(',')
    if mech not in supported_mechs:
        """
        How we get here:

        1. Sopel connects to a network advertising SASL 3.1
        2. SASL 3.1 doesn't advertise supported mechanisms up front, so Sopel
           blindly goes ahead with whatever SASL config it's set to use
        3. The server doesn't support the mechanism Sopel used, and is a good
           IRC citizen, so it sends this optional numeric, 908 RPL_SASLMECHS

        Note that misconfigured SASL 3.1 will just silently fail when connected
        to an IRC server NOT implementing the optional 908 reply.

        A network with SASL 3.2 should theoretically never get this far because
        Sopel should catch the unadvertised mechanism in CAP_SASL.

        See https://github.com/sopel-irc/sopel/issues/1780 for background
        """
        LOGGER.error(
            "Configured SASL mechanism '%s' is not advertised by this server. "
            "Advertised values: %s",
            mech,
            ', '.join(supported_mechs),
        )
        bot.resume_capability_negotiation(CAP_SASL.cap_req, 'coretasks')
        bot.quit('Wrong SASL configuration.')
    else:
        LOGGER.info(
            "Selected SASL mechanism is %s, advertised: %s",
            mech,
            ', '.join(supported_mechs),
        )


def _get_sasl_pass_and_mech(bot):
    password = None
    mech = None

    if bot.config.core.auth_method == 'sasl':
        password = bot.config.core.auth_password
        mech = bot.config.core.auth_target
    elif bot.config.core.server_auth_method == 'sasl':
        password = bot.config.core.server_auth_password
        mech = bot.config.core.server_auth_sasl_mech

    mech = 'PLAIN' if mech is None else mech.upper()

    return password, mech


# Live blocklist editing


@plugin.commands('blocks')
@plugin.example(r'.blocks del nick falsep0sitive', user_help=True)
@plugin.example(r'.blocks add host some\.malicious\.network', user_help=True)
@plugin.example(r'.blocks add nick sp(a|4)mb(o|0)t\d*', user_help=True)
@plugin.thread(False)
@plugin.unblockable
@plugin.priority('low')
@plugin.require_admin
def blocks(bot, trigger):
    """Manage Sopel's blocking features.

    Full argspec: `list [nick|host]` or `[add|del] [nick|host] pattern`
    """
    STRINGS = {
        "success_del": "Successfully deleted block: %s",
        "success_add": "Successfully added block: %s",
        "no_nick": "No matching nick block found for: %s",
        "no_host": "No matching host block found for: %s",
        "invalid": "Invalid format for %s a block. Try: .blocks add (nick|host) sopel",
        "invalid_display": "Invalid input for displaying blocks.",
        "nonelisted": "No %s listed in the blocklist.",
        'huh': "I could not figure out what you wanted to do.",
    }

    hosts = set(s for s in bot.config.core.host_blocks if s != '')
    nicks = set(bot.make_identifier(nick)
                for nick in bot.config.core.nick_blocks
                if nick != '')
    text = trigger.group().split()

    if len(text) == 3 and text[1] == "list":
        if text[2] == "host":
            if len(hosts) > 0:
                blocked = ', '.join(str(host) for host in hosts)
                bot.say("Blocked hosts: {}".format(blocked))
            else:
                bot.reply(STRINGS['nonelisted'] % ('hosts'))
        elif text[2] == "nick":
            if len(nicks) > 0:
                blocked = ', '.join(str(nick) for nick in nicks)
                bot.say("Blocked nicks: {}".format(blocked))
            else:
                bot.reply(STRINGS['nonelisted'] % ('nicks'))
        else:
            bot.reply(STRINGS['invalid_display'])

    elif len(text) == 4 and text[1] == "add":
        if text[2] == "nick":
            nicks.add(text[3])
            bot.config.core.nick_blocks = nicks
            bot.config.save()
        elif text[2] == "host":
            hosts.add(text[3].lower())
            bot.config.core.host_blocks = list(hosts)
            bot.config.save()
        else:
            bot.reply(STRINGS['invalid'] % ("adding"))
            return

        bot.reply(STRINGS['success_add'] % (text[3]))

    elif len(text) == 4 and text[1] == "del":
        if text[2] == "nick":
            nick = bot.make_identifier(text[3])
            if nick not in nicks:
                bot.reply(STRINGS['no_nick'] % (text[3]))
                return
            nicks.remove(nick)
            bot.config.core.nick_blocks = [str(n) for n in nicks]
            bot.config.save()
            bot.reply(STRINGS['success_del'] % (text[3]))
        elif text[2] == "host":
            host = text[3].lower()
            if host not in hosts:
                bot.reply(STRINGS['no_host'] % (text[3]))
                return
            hosts.remove(host)
            bot.config.core.host_blocks = [str(m) for m in hosts]
            bot.config.save()
            bot.reply(STRINGS['success_del'] % (text[3]))
        else:
            bot.reply(STRINGS['invalid'] % ("deleting"))
            return
    else:
        bot.reply(STRINGS['huh'])


@plugin.event('CHGHOST')
@plugin.thread(False)
@plugin.unblockable
@plugin.priority('medium')
def recv_chghost(bot, trigger):
    """Track user/host changes."""
    if trigger.nick not in bot.users:
        bot.users[trigger.nick] = target.User(
            trigger.nick, trigger.user, trigger.host)

    try:
        new_user, new_host = trigger.args
    except ValueError:
        LOGGER.warning(
            "Ignoring CHGHOST command with %s arguments: %r",
            'extra' if len(trigger.args) > 2 else 'insufficient',
            trigger.args)
        return

    bot.users[trigger.nick].user = new_user
    bot.users[trigger.nick].host = new_host
    LOGGER.info(
        "Update user@host for nick %r: %s@%s",
        str(trigger.nick), new_user, new_host)


@plugin.event('ACCOUNT')
@plugin.thread(False)
@plugin.unblockable
@plugin.priority('medium')
def account_notify(bot, trigger):
    """Track users' accounts."""
    if trigger.nick not in bot.users:
        bot.users[trigger.nick] = target.User(
            trigger.nick, trigger.user, trigger.host)
    account = trigger.args[0]
    if account == '*':
        account = None
    bot.users[trigger.nick].account = account
    LOGGER.info("Update account for nick %r: %s", str(trigger.nick), account)


@plugin.event(events.RPL_WHOSPCRPL)
@plugin.thread(False)
@plugin.unblockable
@plugin.priority('medium')
def recv_whox(bot, trigger):
    """Track ``WHO`` responses when ``WHOX`` is enabled."""
    if len(trigger.args) < 2 or trigger.args[1] != WHOX_QUERYTYPE:
        # Ignored, some plugin probably called WHO
        LOGGER.debug("Ignoring WHO reply for channel '%s'; not queried by coretasks", trigger.args[1])
        return
    if len(trigger.args) != len(WHOX_QUERY):
        LOGGER.warning(
            "While populating `bot.accounts` a WHO response was malformed.")
        return
    _, _, channel, user, host, nick, status, account, realname = trigger.args
    botmode = bot.isupport.get('BOT')
    away = 'G' in status
    is_bot = (botmode in status) if botmode else None
    modes = ''.join([c for c in status if c in '~&@%+!'])
    _record_who(bot, channel, user, host, nick, realname, account, away, is_bot, modes)


def _record_who(
    bot: Sopel,
    channel: Identifier,
    user: str,
    host: str,
    nick: str,
    realname: Optional[str] = None,
    account: Optional[str] = None,
    away: Optional[bool] = None,
    is_bot: Optional[bool] = None,
    modes: Optional[str] = None,
) -> None:
    nick = bot.make_identifier(nick)
    channel = bot.make_identifier(channel)
    if nick not in bot.users:
        usr = target.User(nick, user, host)
        bot.users[nick] = usr
    else:
        usr = bot.users[nick]
        # check for & fill in sparse User added by handle_names()
        if usr.host is None and host:
            usr.host = host
        if usr.user is None and user:
            usr.user = user
    if realname:
        usr.realname = realname
    if account == '0':
        usr.account = None
    else:
        usr.account = account
    if away is not None:
        usr.away = away
    if is_bot is not None:
        usr.is_bot = is_bot
    priv = 0
    if modes:
        mapping = {
            "+": plugin.VOICE,
            "%": plugin.HALFOP,
            "@": plugin.OP,
            "&": plugin.ADMIN,
            "~": plugin.OWNER,
            "!": plugin.OPER,
        }
        for c in modes:
            priv = priv | mapping[c]
    if channel not in bot.channels:
        bot.channels[channel] = target.Channel(
            channel,
            identifier_factory=bot.make_identifier,
        )

    bot.channels[channel].add_user(usr, privs=priv)


@plugin.event(events.RPL_WHOREPLY)
@plugin.thread(False)
@plugin.unblockable
@plugin.priority('medium')
def recv_who(bot, trigger):
    """Track ``WHO`` responses when ``WHOX`` is not enabled."""
    channel, user, host, _, nick, status = trigger.args[1:7]
    botmode = bot.isupport.get('BOT')
    realname = trigger.args[-1].partition(' ')[-1]
    away = 'G' in status
    is_bot = (botmode in status) if botmode else None
    modes = ''.join([c for c in status if c in '~&@%+!'])
    _record_who(
        bot, channel, user, host, nick, realname,
        away=away, is_bot=is_bot, modes=modes,
    )


@plugin.event('AWAY')
@plugin.thread(False)
@plugin.unblockable
@plugin.priority('medium')
def track_notify(bot, trigger):
    """Track users going away or coming back."""
    if trigger.nick not in bot.users:
        bot.users[trigger.nick] = target.User(
            trigger.nick, trigger.user, trigger.host)
    user = bot.users[trigger.nick]
    user.away = bool(trigger.args)
    state_change = 'went away' if user.away else 'came back'
    LOGGER.info("User %s: %s", state_change, trigger.nick)


@plugin.event('TOPIC')
@plugin.event(events.RPL_TOPIC)
@plugin.thread(False)
@plugin.unblockable
@plugin.priority('medium')
def track_topic(bot, trigger):
    """Track channels' topics."""
    if trigger.event != 'TOPIC':
        channel = trigger.args[1]
    else:
        channel = trigger.args[0]
    if channel not in bot.channels:
        return
    bot.channels[channel].topic = trigger.args[-1]
    LOGGER.info("Channel's topic updated: %s", channel)


@plugin.rule(r'(?u).*(.+://\S+).*')
def handle_url_callbacks(bot, trigger):
    """Dispatch callbacks on URLs

    For each URL found in the trigger, trigger the URL callback registered by
    the ``@url`` decorator.
    """
    # find URLs in the trigger
    for url in trigger.urls:
        # find callbacks for said URL
        for function, match in bot.search_url_callbacks(url):
            # trigger callback defined by the `@url` decorator
            if hasattr(function, 'url_regex'):
                # bake the `match` argument in before passing the callback on
                @functools.wraps(function)
                def decorated(bot, trigger):
                    return function(bot, trigger, match=match)

                bot.call(decorated, bot, trigger)
